/*********************************************************
 *
 * This source code is part of the Carnegie Mellon Robot
 * Navigation Toolkit (CARMEN)
 *
 * CARMEN Copyright (c) 2002 Michael Montemerlo, Nicholas
 * Roy, and Sebastian Thrun
 *
 * CARMEN is free software; you can redistribute it and/or 
 * modify it under the terms of the GNU General Public 
 * License as published by the Free Software Foundation; 
 * either version 2 of the License, or (at your option)
 * any later version.
 *
 * CARMEN is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied 
 * warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
 * PURPOSE.  See the GNU General Public License for more 
 * details.
 *
 * You should have received a copy of the GNU General 
 * Public License along with CARMEN; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place, 
 * Suite 330, Boston, MA  02111-1307 USA
 *
 ********************************************************/

#include <carmen/carmen_graphics.h>
#include <gdk/gdkx.h>
#include <X11/Xlib.h>
#include <X11/extensions/shape.h>

#define WINDOWSIZE 500

static carmen_laser_laser_message laser;
static int received_laser = 0;
static int laser_num = 0;
static GtkWidget *drawing_area;
static double laser_range = 5.0;
static int laser_count = 0;
static double start_time = 0.0;

static void Redraw(void);

static void 
laser_handler(void)
{
  received_laser = 1;
  laser_count++;
  Redraw();
}

static void 
shutdown_laserview(int x)
{
  if(x == SIGINT) {
    close_ipc();
    exit(1);
  }
}

static gint 
updateIPC(gpointer *data __attribute__ ((unused))) 
{
  sleep_ipc(0.01);
  carmen_graphics_update_ipc_callbacks((GdkInputFunction)updateIPC);
  return 1;
}

static gint 
Expose_Event(GtkWidget *widget __attribute__ ((unused)), 
	     GdkEventExpose *event __attribute__ ((unused))) 
{
  Redraw();
  return 1;
}

static void 
Redraw(void)
{
  float origin_x, origin_y, x = 0, y = 0, last_x, last_y, angle, scale;
  int i;
  static GdkGC *Drawing_GC = NULL;
  static GdkPixmap *pixmap = NULL;
  GdkPoint *poly = NULL;
  char str[20];
  static double framerate = 0.0;

  if(Drawing_GC == NULL) {
    carmen_graphics_setup_colors();
    Drawing_GC = gdk_gc_new(drawing_area->window);
  }

  /* setup graphics parameters */
  if(pixmap == NULL)
    pixmap = gdk_pixmap_new(drawing_area->window, 
			    drawing_area->allocation.width,
			    drawing_area->allocation.height, -1);
  if(pixmap == NULL)
    return;

  /* erase window with light blue */
  gdk_gc_set_foreground(Drawing_GC, &carmen_light_blue);
  gdk_draw_rectangle(pixmap, Drawing_GC, TRUE, 0, 0, 
		     drawing_area->allocation.width, 
		     drawing_area->allocation.height);

  gdk_gc_set_line_attributes(Drawing_GC, 4, GDK_LINE_SOLID,
                             GDK_CAP_NOT_LAST, GDK_JOIN_MITER);
  gdk_gc_set_foreground(Drawing_GC, &carmen_blue);
  gdk_draw_arc(pixmap, Drawing_GC, FALSE, 0, 0, 
	       drawing_area->allocation.width,
	       drawing_area->allocation.height * 2, 0, 180 * 64);
  gdk_draw_line(pixmap, Drawing_GC, 
		0, drawing_area->allocation.height,
		drawing_area->allocation.width, 
		drawing_area->allocation.height);
  gdk_gc_set_line_attributes(Drawing_GC, 1, GDK_LINE_SOLID,
                             GDK_CAP_NOT_LAST, GDK_JOIN_MITER);

  if(received_laser) {
    /* draw the empty space */
    poly = (GdkPoint *)realloc(poly, sizeof(GdkPoint) * 
			       (laser.num_readings + 1));
    carmen_test_alloc(poly);
    origin_x = drawing_area->allocation.width / 2; 
    origin_y = drawing_area->allocation.height - 1;
    scale = drawing_area->allocation.height / (double)laser_range;
    for(i = 0; i < laser.num_readings; i++) {
      angle = i / (float)(laser.num_readings - 1) * M_PI;
      x = laser.range[i] * cos(angle);
      y = laser.range[i] * sin(angle);
      poly[i].x = origin_x + x * scale;
      poly[i].y = origin_y - y * scale;
    }
    poly[laser.num_readings].x = origin_x;
    poly[laser.num_readings].y = origin_y;
    gdk_gc_set_foreground(Drawing_GC, &carmen_white);
    gdk_draw_polygon(pixmap, Drawing_GC, TRUE, poly, 
		     laser.num_readings + 1);

    /* connect nearby laser points with lines */
    gdk_gc_set_line_attributes(Drawing_GC, 2, GDK_LINE_SOLID,
			       GDK_CAP_NOT_LAST, GDK_JOIN_MITER);
    gdk_gc_set_foreground(Drawing_GC, &carmen_blue);
    for(i = 0; i < laser.num_readings; i++) {
      last_x = x;
      last_y = y;
      angle = i / (float)(laser.num_readings - 1) * M_PI;
      x = laser.range[i] * cos(angle);
      y = laser.range[i] * sin(angle);
      if(i > 0 && hypot(x - last_x, y - last_y) < .20)
	gdk_draw_line(pixmap, Drawing_GC, 
		      origin_x + last_x * scale,
		      origin_y - last_y * scale,
		      origin_x + x * scale,
		      origin_y - y * scale);
    }

    free(poly);
  }

  if(laser_count % 10 == 0 || laser_count < 10)
    framerate = laser_count / (carmen_get_time_ms() - start_time);

  sprintf(str, "Laser %d - %.1f fps", laser_num, framerate);
  gdk_gc_set_foreground(Drawing_GC, &carmen_black);
  gdk_draw_string(pixmap, drawing_area->style->font, Drawing_GC, 
		  10, drawing_area->allocation.height - 10, str);

  sprintf(str, "%d pts", laser.num_readings);
  gdk_gc_set_foreground(Drawing_GC, &carmen_black);
  gdk_draw_string(pixmap, drawing_area->style->font, Drawing_GC, 
		  drawing_area->allocation.width - 60.0, 
		  drawing_area->allocation.height - 25, str);

  sprintf(str, "Max range  = %.1fm", laser_range);
  gdk_gc_set_foreground(Drawing_GC, &carmen_black);
  gdk_draw_string(pixmap, drawing_area->style->font, Drawing_GC, 
		  drawing_area->allocation.width - 125.0, 
		  drawing_area->allocation.height - 10, str);
  
  /* udpate the whole window */
  gdk_draw_pixmap(drawing_area->window, 
		  drawing_area->style->fg_gc[GTK_WIDGET_STATE (drawing_area)],
                  pixmap, 0, 0, 0, 0, 
		  drawing_area->allocation.width, 
		  drawing_area->allocation.height);
}

static void 
start_graphics(int argc, char *argv[]) 
{
  GtkWidget *main_window;
  Pixmap mask;
  int r, c, byte, bit, rowspan;
  Display *display;
  Window window;
  unsigned char data[WINDOWSIZE * WINDOWSIZE];

  gtk_init(&argc, &argv);
  carmen_graphics_initialize_screenshot();
  main_window = gtk_window_new (GTK_WINDOW_TOPLEVEL);
  gtk_window_set_title (GTK_WINDOW (main_window), "Robot Graph");
  drawing_area = gtk_drawing_area_new ();
  gtk_widget_set_usize (drawing_area, WINDOWSIZE, WINDOWSIZE / 2);
  gtk_container_add(GTK_CONTAINER(main_window), drawing_area);
  gtk_signal_connect(GTK_OBJECT(drawing_area), "expose_event",
		     (GtkSignalFunc)Expose_Event, NULL);
  gtk_widget_add_events(drawing_area,  GDK_EXPOSURE_MASK);
  carmen_graphics_update_ipc_callbacks((GdkInputFunction)updateIPC);

  gtk_widget_realize(main_window);
  //  gdk_window_set_decorations(main_window->window, GDK_DECOR_TITLE);
  gdk_window_set_decorations(main_window->window, 0);
  gdk_window_set_functions(main_window->window, 0);
  
  gtk_widget_show(drawing_area);
  gtk_widget_show(main_window);

  rowspan = ceil(WINDOWSIZE / 8.0);
  for(r = 0; r < WINDOWSIZE / 2; r++)
    for(c = 0; c < WINDOWSIZE; c++) {
      byte = r * rowspan + c / 8;
      bit =  c % 8;
      if(hypot(WINDOWSIZE / 2 - r, WINDOWSIZE / 2 - c) < WINDOWSIZE / 2)
	data[byte] = data[byte] | (1 << (bit));
      else
	data[byte] = data[byte] & (~(1 << (bit)));
    }
  display = GDK_WINDOW_XDISPLAY(main_window->window);
  window = GDK_WINDOW_XWINDOW(main_window->window);
  mask = XCreatePixmapFromBitmapData(display, window, data,
				     WINDOWSIZE, WINDOWSIZE, 1, 0, 1);
  XShapeCombineMask(display, window,
		    ShapeBounding, 0, 0, mask, ShapeIntersect);
  gtk_main();
}

int 
main(int argc, char **argv)
{  
  carmen_initialize_ipc(argv[0]);
  carmen_param_check_version(argv[0]);
  if(argc >= 2)
    laser_num = atoi(argv[1]);
  if(argc >=3)
    laser_range = atof(argv[2]);

  if(laser_num == 0)
    carmen_laser_subscribe_frontlaser_message(&laser,
					      (carmen_handler_t)
					      laser_handler,
					      CARMEN_SUBSCRIBE_LATEST);
  else
    carmen_laser_subscribe_rearlaser_message(&laser,
					     (carmen_handler_t)
					     laser_handler,
					     CARMEN_SUBSCRIBE_LATEST);
  signal(SIGINT, shutdown_laserview);
  start_time = carmen_get_time_ms();
  start_graphics(argc, argv);
  return 0;
}
