/*********************************************************
 *
 * This source code is part of the Carnegie Mellon Robot
 * Navigation Toolkit (CARMEN)
 *
 * CARMEN Copyright (c) 2002 Michael Montemerlo, Nicholas
 * Roy, and Sebastian Thrun
 *
 * CARMEN is free software; you can redistribute it and/or 
 * modify it under the terms of the GNU General Public 
 * License as published by the Free Software Foundation; 
 * either version 2 of the License, or (at your option)
 * any later version.
 *
 * CARMEN is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied 
 * warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
 * PURPOSE.  See the GNU General Public License for more 
 * details.
 *
 * You should have received a copy of the GNU General 
 * Public License along with CARMEN; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place, 
 * Suite 330, Boston, MA  02111-1307 USA
 *
 ********************************************************/

/***********************************
 * map_editor_graphics deals with  *
 * the pixmaps and the main window *
 * which display the map.          *
 ***********************************/
#include <gnome.h>
#include <carmen/carmen_graphics.h>

#include "map_editor.h"
#include "map_editor_graphics.h"
#include "map_editor_drawing.h"
#include "map_editor_menus.h"

#define DEFAULT_MAX_SCREEN_WIDTH 600
#define DEFAULT_MAX_SCREEN_HEIGHT 400

static GtkObject *x_adjustment, *y_adjustment;
static GtkWidget *drawing_table, *window_box;

static GdkFont *place_font;
static int current_place = -1;
GtkItemFactory *item_factory;

/*****************************************************************
 * Conversions from the coordinates on the screen to the         *
 * coordinates on the map and vis versa                          *
 *****************************************************************/

/* converts an x position on the drawing area to an x grid
   on the map */
inline double 
pix_x_to_map(double pix_x)
{
  return (double)(pix_x)/mult + (double)xstart;
}

/* converts a y position on the drawing area to a y grid
   on the map */
inline double 
pix_y_to_map(double pix_y)
{
  return (double)yend - (double)(pix_y)/mult;
}
 
/* converts an x grid to an x position on the drawing area */
inline double 
map_x_to_pix(int map_x)
{
  return (double)(map_x-xstart)*mult;
}

/* converts a y grid to a y position on the drawing area */
inline double 
map_y_to_pix(int map_y)
{
  return (double)(yend-map_y)*mult;
}

/* converts an x grid to a position on the map_pixmap */
inline double 
map_x_to_map_pix(int map_x)
{
  return (double)(map_x)*mult;
}

/* converts an x grid to a position on the map_pixmap */
inline double 
map_y_to_map_pix(int map_y)
{
  return (double)(map->config.y_size-map_y)*mult;
}


/*****************************************************************
 * Grafics helper functions                                      *
 *****************************************************************/

void 
draw_offlimits(int i)
{
  int pix_x1, pix_y1, pix_x2, pix_y2;

  pix_x1 = map_x_to_pix(offlimits_array[i].x1);
  pix_y1 = map_y_to_pix(offlimits_array[i].y1);
  pix_x2 = map_x_to_pix(offlimits_array[i].x2);
  pix_y2 = map_y_to_pix(offlimits_array[i].y2);
	
  switch (offlimits_array[i].type) {
  case CARMEN_OFFLIMITS_POINT_ID:
    break;
  case CARMEN_OFFLIMITS_LINE_ID:
    gdk_draw_line (map_pixmap, Drawing_GC, pix_x1, pix_y1, pix_x2, 
		   pix_y2);
    break;
  case CARMEN_OFFLIMITS_RECT_ID:
    gdk_draw_rectangle(map_pixmap, Drawing_GC, 1, pix_x1, pix_y2, 
		       pix_x2-pix_x1, pix_y1-pix_y2);
    break;
  default:
    break;
  }
}

void 
draw_place_name(int i, GdkPixmap *pixmap, GdkColor *draw_color)
{
  int pix_x1, pix_y1;

  pix_x1 = map_x_to_pix(place_list->places[i].x/map->config.resolution);
  pix_y1 = map_y_to_pix(place_list->places[i].y/map->config.resolution);
	
  gdk_gc_set_foreground (Drawing_GC, draw_color);

  gdk_draw_arc(pixmap, Drawing_GC, 1, pix_x1-5, pix_y1-5,
	       10, 10, 0, 360*64);

  gdk_gc_set_foreground (Drawing_GC, &carmen_black);

  gdk_draw_arc(pixmap, Drawing_GC, 0, pix_x1-5, pix_y1-5,
	       10, 10, 0, 360*64); 

  gdk_gc_set_foreground (Drawing_GC, &carmen_black);
  gdk_draw_string(pixmap, place_font, Drawing_GC, pix_x1, pix_y1-5,
		  place_list->places[i].name);
}

/* draws the map_pixmap onto the tmp_pixmap */
void 
map_to_tmp(void)
{
  int xsrc, ysrc, xdest, ydest, w,h;
  if (xstart >= 0)
    {
      xsrc = xstart*mult;
      xdest = 0;
      if(xend < map->config.x_size)
	w = drawing_area->allocation.width;
      else
	w = (map->config.x_size-xstart)*mult;
    }
  else
    {
      xsrc = 0;
      xdest = -xstart*mult;
      if(xend < map->config.x_size)
	w = drawing_area->allocation.width + xstart*mult;
      else
	w = map->config.x_size*mult;
    }
  if (yend < map->config.y_size)
    {
      ysrc = (map->config.y_size-yend)*mult;
      ydest = 0;
      if(ystart >= 0)
	h = drawing_area->allocation.height;
      else
	h = yend*mult;
    }
  else
    {
      ysrc = 0;
      ydest = (yend-map->config.y_size)*mult;
      if (ystart >= 0)
	h = drawing_area->allocation.height+(map->config.y_size-yend)*mult;
      else
	h = map->config.y_size*mult;
    }

  gdk_gc_set_foreground(Drawing_GC, &purple);
  gdk_draw_rectangle(tmp_pixmap, Drawing_GC, TRUE, 0, 0,
		     drawing_area->allocation.width,
		     drawing_area->allocation.height);

  gdk_draw_pixmap(tmp_pixmap, 
		  drawing_area->style->fg_gc[GTK_WIDGET_STATE (drawing_area)],
		  map_pixmap, xsrc, ysrc, xdest, ydest, w, h);
}


/* sets up the default colors and Drawing_GC */
void 
setup_colors(void)
{
  if (Drawing_GC == NULL) 
    Drawing_GC = gdk_gc_new(drawing_area->window);
  if (Drawing_GC == NULL)	
    carmen_die("Drawing_GC could not be initialized\n");

  yellow = carmen_graphics_add_color("Yellow");
  blue = carmen_graphics_add_color_rgb(0, 0, 255);
  red = carmen_graphics_add_color_rgb(255, 0, 0);
  purple = carmen_graphics_add_color_rgb(150, 0, 150);
}

/* redraws the pixmap onto the drawing area */
void 
redraw(void)
{
  int i;

  if(map_pixmap == NULL)
    {
      if (image_data != NULL)
	free(image_data);
      setup_colors();
      image_data = carmen_graphics_convert_to_image(map, 0);
      map_pixmap = carmen_graphics_generate_pixmap(drawing_area, image_data, 
						   &(map->config), mult);

      if (show_offlimits) 
	{
	  gdk_gc_set_foreground (Drawing_GC, &red);
	  for (i = 0; i < num_offlimits_segments; i++) 
	    draw_offlimits(i);
	}

      if (show_place_names && place_list)
	{
	  for (i = 0; i < place_list->num_places; i++) 
	    draw_place_name(i, map_pixmap, &carmen_red);
	}

      tmp_pixmap = gdk_pixmap_new(drawing_area->window, 
				  drawing_area->allocation.width,
				  drawing_area->allocation.height, -1);
      map_to_tmp();
    }	
  
  gdk_draw_pixmap(drawing_area->window,   //dest
		  drawing_area->style->fg_gc[GTK_WIDGET_STATE (drawing_area)],
		  tmp_pixmap,             //src
		  0, 0, 0, 0, -1, -1);

}

/***************************************************************************
 * Event handlers for the drawing area                                     *
 ***************************************************************************/

/* handles a button press event in the drawing area.
   calls the apropriate tool depending upon utensil's setting 
   also creates a backup for undo and incrememnts modified if necessary */
static gint 
button_press_event( GtkWidget *widget, GdkEventButton *event )
{  
  if(map_pixmap == NULL)
    return -1;

  if (deleting_placename)
    return TRUE;

  if (adding_placename)
    return TRUE;

  if (event->button == 1)
    {
      switch(utensil)
	{
	case CARMEN_MAP_BRUSH:
	  memcpy(backup, map->complete_map, sizeof(float) * map->config.x_size * 
		 map->config.y_size); 
	  modified ++;
	  draw_brush(widget, event->x, event->y);
	  break;
	case CARMEN_MAP_RECTANGLE:
	  memcpy(backup, map->complete_map, sizeof(float) * map->config.x_size * 
		 map->config.y_size);
	  modified ++;
	  creating_rectangle(widget, 0, event->x, event->y);
	  break;
	case CARMEN_MAP_LINE:
	  memcpy(backup, map->complete_map, sizeof(float) * map->config.x_size * 
		 map->config.y_size);
	  modified ++;
	  creating_line(0, event->x, event->y);
	  break;
	case CARMEN_MAP_FILL:
	  memcpy(backup, map->complete_map, sizeof(float) * map->config.x_size * 
		 map->config.y_size);
	  modified ++;
	  create_fill(event->x, event->y);
	  break;
	case CARMEN_MAP_FUZZY_FILL:
	  memcpy(backup, map->complete_map, sizeof(float) * map->config.x_size * 
		 map->config.y_size);
	  modified ++;
	  create_fuzzy_fill(event->x, event->y);
	  break;
	case CARMEN_MAP_EYEDROPPER:
	  sample(event->x, event->y);
	  break;
	case CARMEN_MAP_ZOOM:
	  zoom_in(event->x, event->y);
	  break;
	}
    }
  else if (event->button == 3 && utensil == CARMEN_MAP_ZOOM)
    zoom_out(event->x, event->y);

  redraw();
  return TRUE;
}

/* handles the mouse button being released inside the drawing area 
   calls the apropriate tool function, or none */
static gint 
button_release_event(GtkWidget *widget, GdkEventButton *event )
{
  double x, y;

  if (deleting_placename) {
    do_delete_placename(current_place);
    return TRUE;
  }
  if (adding_placename) {
    x = pix_x_to_map(event->x)*map->config.resolution;
    y = pix_y_to_map(event->y)*map->config.resolution;
    start_add_placename(x, y);
    return TRUE;
  }

  if (event->button == 1 && map_pixmap != NULL)
    switch(utensil)
      {
      case CARMEN_MAP_RECTANGLE:
	creating_rectangle(widget, 2, event->x, event->y);
	break;
      case CARMEN_MAP_LINE:
	creating_line(2, event->x, event->y);
	break;
      default:
	break;
      }

  redraw();
  return 1;
}

static void
handle_deleting_placename_move(int map_x, int map_y)
{
  double closest_distance;
  int closest_place;

  double distance;
  int i;

  closest_distance = hypot(map_x-place_list->places[0].x/
			   map->config.resolution,
			   map_y-place_list->places[0].y/
			   map->config.resolution);
  closest_place = 0;
  for (i = 0; i < place_list->num_places; i++) 
    {
      distance = hypot(map_x-place_list->places[i].x/
		       map->config.resolution,
		       map_y-place_list->places[i].y/
		       map->config.resolution);
      if (distance < closest_distance) 
	{
	  closest_distance = distance;
	  closest_place = i;
	}
    }

  if (closest_place != current_place && current_place >= 0)
    draw_place_name(current_place, tmp_pixmap, &carmen_red);

  current_place = closest_place;
  draw_place_name(closest_place, tmp_pixmap, &carmen_yellow);

}


/* handles the mouse moving in the drawing area
   only calls the apropriate tool function if the mouse button is down.
   always sets the cursor to the apropriate simbol */
static gint 
motion_notify_event( GtkWidget *widget, GdkEventMotion *event )
{
  int x, y;
  GdkModifierType state;
  static char label_buffer[255];
  static carmen_map_point_t map_point;
  static carmen_world_point_t world_point;
	

  GTK_WIDGET_CLASS(GTK_OBJECT(vrule)->klass)->motion_notify_event(vrule, event);
  GTK_WIDGET_CLASS(GTK_OBJECT(hrule)->klass)->motion_notify_event(hrule, event);

  //   map_to_tmp();

  if (event->is_hint)
    gdk_window_get_pointer (event->window, &x, &y, &state);
  else
    {
      x = event->x;
      y = event->y;
      state = event->state;
    }
	
  map_point.x = pix_x_to_map(x);
  map_point.y = pix_y_to_map(y);
  map_point.map = map;
  carmen_map_to_world(&map_point, &world_point);
  sprintf(label_buffer, "X: %6.2f Y: %6.2f", world_point.pose.x, 
	  world_point.pose.y);
  gtk_label_set_text(GTK_LABEL(coords_label), label_buffer);

  if (map_pixmap == NULL)
    return TRUE;

  if (deleting_placename) 
    {
      handle_deleting_placename_move(map_point.x, map_point.y);
    }

  else if ((state & GDK_BUTTON1_MASK)) 
    {
      switch(utensil)
	{
	case CARMEN_MAP_BRUSH:
	  draw_brush (widget, x, y);
	  x = x/mult - brush_size + .5;
	  y = y/mult - brush_size + .5;
	  break;
	case CARMEN_MAP_RECTANGLE:
	  creating_rectangle(widget, 1, x, y);
	  break;
	case CARMEN_MAP_LINE:
	  creating_line(1, x, y);
	  break;
	case CARMEN_MAP_EYEDROPPER:
	  sample(x, y);
	  break;
	default:
	  break;
	}
    }

  redraw();
  
  return TRUE;
}

static gint 
expose_event(GtkWidget *widget __attribute__ ((unused)), 
	     GdkEventExpose *event __attribute__ ((unused)))
{
  redraw();
  return FALSE;
}

/*************************************************************************
 * event handlers for the scroll bars                                    *
 *************************************************************************/

gint 
horizontal(GtkAdjustment *adj)
{
  GtkAdjustment *adjustment;
  int delta;
  int map_x;

  map_x = adj->value;
  delta = (xend-xstart);
  xstart = adj->value;
  xend = xstart + delta;

  gtk_ruler_set_range(GTK_RULER(hrule),
		      map->config.resolution*xstart/100.0, 
		      map->config.resolution*xend/100.0, 
		      map->config.resolution * map_x / 100.0, 
		      map->config.resolution * map->config.x_size / 100.0);

  adjustment = gtk_range_get_adjustment(GTK_RANGE(hscroll));
  if(adjustment->lower < 0 && adjustment->lower < xstart)
    {
      if(xstart > 0)
	adjustment->lower = 0;
      else
	adjustment->lower = xstart;
      gtk_signal_emit_by_name (GTK_OBJECT (adjustment), "changed");
    }
  if(adjustment->upper > map->config.x_size && adjustment->upper > xend)
    {
      if(xend < map->config.x_size)
	adjustment->upper = map->config.x_size;
      else
	adjustment->upper = xend;
      gtk_signal_emit_by_name (GTK_OBJECT (adjustment), "changed");
    }
  if(tmp_pixmap)
    {
      map_to_tmp();
      redraw();
    }
  
  return 1;
}

gint 
vertical(GtkAdjustment *adj)
{
  GtkAdjustment *adjustment;
  int delta;
  int map_y;

  map_y = map->config.y_size - adj->value;
  delta = (yend-ystart);
  yend = map->config.y_size - adj->value;
  ystart = yend - delta;;

  gtk_ruler_set_range(GTK_RULER(vrule),
		      map->config.resolution*yend/100.0,  
		      map->config.resolution*ystart/100.0,
		      map->config.resolution * map_y / 100.0, 
		      map->config.resolution * map->config.y_size / 100.0);

  adjustment = gtk_range_get_adjustment(GTK_RANGE(vscroll));
  if(map->config.y_size - adjustment->upper < 0 && map->config.y_size - adjustment->upper < ystart)
    {
      if(ystart > 0)
	adjustment->upper = map->config.y_size;
      else
	adjustment->upper = map->config.y_size - ystart;
      gtk_signal_emit_by_name (GTK_OBJECT (adjustment), "changed");
    }
  if(map->config.y_size - adjustment->lower > map->config.y_size && map->config.y_size - adjustment->lower > yend)
    {
      if(yend < map->config.y_size)
	adjustment->lower = 0;
      else
	adjustment->lower = map->config.y_size - yend;
      gtk_signal_emit_by_name (GTK_OBJECT (adjustment), "changed");
    }
    
  if(tmp_pixmap)
    {
      map_to_tmp();
      redraw();
    }
  
  return 1;
}

/*************************************************************************
 * event handlers for the main window                                    *
 *************************************************************************/

gint 
off_pixmap(void)
{
  if(tmp_pixmap)
    {
      map_to_tmp();
      redraw();
    }
  fprintf(stderr, "off pixmap\n");
  set_point();
  return 1;
}

/*************************************************************************
 * startup                                                               *
 *************************************************************************/

static GtkItemFactoryEntry menu_items[] = {
  { "/_File",         NULL,         NULL, 0, "<Branch>" },
  { "/File/_New Map...",    "<control>N", new_map_menu, 0, NULL },
  { "/File/_Open Map...",    "<control>O", open_map_menu, 0, NULL },
  { "/File/_Save Map",    "<control>S", save_map_menu, 0, NULL },
  { "/File/Save Map _As...", NULL,         save_map_as_menu, 0, NULL },
  { "/File/_Export...", NULL,         NULL, 0, NULL },
  { "/File/sep1",    NULL,        NULL, 0, "<Separator>" },
  { "/File/_Quit",    "<control>Q", quit_menu, 0, NULL },
  { "/_Edit",         NULL,         NULL, 0, "<Branch>" },
  { "/Edit/_Undo",    "<control>Z",         undo_menu, 0, NULL },
  { "/Edit/sep1",    NULL,         NULL, 0, "<Separator>" },
  //	{ "/Edit/_Minimize",    NULL,         minimize_map, 0, NULL },
  //	{ "/Edit/sep1",    NULL,         NULL, 0, "<Separator>" },
  { "/Edit/_Add Placename", NULL,  add_placename, 0, NULL },
  { "/Edit/_Delete Placename", NULL, delete_placename, 0, NULL },
  { "/_View",         NULL,         NULL, 0, "<Branch>" },
  { "/View/Show _Placenames", NULL, toggle_view, 1, "<ToggleItem>" },
  { "/View/Show _Offlimits",  NULL, toggle_view, 2, "<ToggleItem>" },
  { "/Help",         NULL,         NULL, 0, "<LastBranch>" },
  { "/Help/About",   NULL,         help_menu, 0, NULL },
};

void get_main_menu(GtkWidget  *the_window, GtkWidget **menubar)
{
  GtkAccelGroup *accel_group;
  gint nmenu_items = sizeof(menu_items) / sizeof(menu_items[0]);
  GtkWidget *menu_item;

  accel_group = gtk_accel_group_new();
  item_factory = gtk_item_factory_new(GTK_TYPE_MENU_BAR, "<main>",
				      accel_group);
  gtk_item_factory_create_items(item_factory, nmenu_items, menu_items, NULL);
  gtk_window_add_accel_group(GTK_WINDOW (the_window), accel_group);
  if(menubar)
    *menubar = gtk_item_factory_get_widget(item_factory, "<main>");

  menu_item = gtk_item_factory_get_item(item_factory, "/View/Show Placenames");
  ((struct _GtkCheckMenuItem *)menu_item)->active = show_place_names;

  menu_item = gtk_item_factory_get_item(item_factory, "/View/Show Offlimits");
  ((struct _GtkCheckMenuItem *)menu_item)->active = show_offlimits;
}

void 
set_up_map_widgets(void)
{
  if (!map)
    return;

  gtk_widget_hide(window);

  xend = map->config.x_size;
  yend = map->config.y_size;

  if(map->config.x_size > map->config.y_size) {
    screen_width = DEFAULT_MAX_SCREEN_WIDTH;
    mult = (double)screen_width / (double)map->config.x_size;
    screen_height = (double)map->config.y_size * (double)mult;
  }
  else {
    screen_height = DEFAULT_MAX_SCREEN_HEIGHT;
    mult = (double)screen_height / (double)map->config.y_size;
    screen_width = (double)map->config.x_size * (double)mult;
  }

  gtk_drawing_area_size (GTK_DRAWING_AREA(drawing_area), screen_width, 
			 screen_height);

  gtk_ruler_set_range(GTK_RULER(hrule),
		      map->config.resolution*xstart/100.0, 
		      map->config.resolution*xend/100.0,  
		      0, map->config.resolution * map->config.x_size / 100.0);
  gtk_ruler_set_range(GTK_RULER(vrule),
		      map->config.resolution*yend/100.0,  
		      map->config.resolution*ystart/100.0,
		      0, map->config.resolution * map->config.y_size / 100.0);

  GTK_ADJUSTMENT(x_adjustment)->value = xstart;
  GTK_ADJUSTMENT(x_adjustment)->lower = 0;
  GTK_ADJUSTMENT(x_adjustment)->upper = map->config.x_size;
  GTK_ADJUSTMENT(x_adjustment)->step_increment = 1;
  GTK_ADJUSTMENT(x_adjustment)->page_increment = xend-xstart;
  GTK_ADJUSTMENT(x_adjustment)->page_size = xend-xstart;

  gtk_signal_emit_by_name (GTK_OBJECT (x_adjustment), "changed");

  GTK_ADJUSTMENT(y_adjustment)->value = map->config.y_size-yend;
  GTK_ADJUSTMENT(y_adjustment)->lower = 0;
  GTK_ADJUSTMENT(y_adjustment)->upper = map->config.y_size;
  GTK_ADJUSTMENT(y_adjustment)->step_increment = 1;
  GTK_ADJUSTMENT(y_adjustment)->page_increment = yend-ystart;
  GTK_ADJUSTMENT(y_adjustment)->page_size = yend-ystart;

  gtk_signal_emit_by_name (GTK_OBJECT (y_adjustment), "changed");

  gtk_widget_show_all(window);
}

/* builds the main window and drawing area for the editor */
void 
start_drawing_window(void)
{
  GtkWidget *box2, *box3, *box4, *box5;
  GtkWidget *separator;
  GtkWidget *text;
  GtkWidget *button;
  GtkWidget *menu_bar;
  GtkObject *adjustment;
  char brush_text[10];
  char ink_text[10];
  char line_text[10];
  char fuzzyness_text[10];

  window = gtk_window_new (GTK_WINDOW_TOPLEVEL);
  gtk_window_set_title (GTK_WINDOW (window), "Map");
  gtk_signal_connect (GTK_OBJECT (window), "delete_event", 
		      GTK_SIGNAL_FUNC (gtk_exit), NULL);
  gtk_widget_set_events (window, GDK_POINTER_MOTION_MASK);

  window_box = gtk_vbox_new(FALSE, 0);
  gtk_container_add(GTK_CONTAINER(window), window_box);
  gtk_widget_show(window_box);

  /* create menus */

  get_main_menu(window, &menu_bar);
  gtk_box_pack_start(GTK_BOX(window_box), menu_bar, FALSE, TRUE, 0);
  gtk_widget_show(menu_bar);

  /* tools */

  box2 = gtk_hbox_new(FALSE, 2);
  gtk_box_pack_start(GTK_BOX(window_box), box2, TRUE, TRUE, 0);
  gtk_widget_show(box2);
  
  separator = gtk_vseparator_new();
  gtk_box_pack_start(GTK_BOX(box2), separator, FALSE, FALSE, 0);
  gtk_widget_show(separator);

  text = gtk_label_new("Tools:");
  gtk_label_set_pattern(GTK_LABEL(text), "_____ ");
  gtk_box_pack_start(GTK_BOX(box2), text, FALSE, FALSE, 0);
  gtk_widget_show(text);

  button = gtk_button_new_with_label ("brush");
  gtk_signal_connect (GTK_OBJECT (button), "clicked", GTK_SIGNAL_FUNC (set_brush), NULL);
  gtk_box_pack_start(GTK_BOX(box2), button, FALSE, FALSE, 0);
  gtk_widget_show(button);

  button = gtk_button_new_with_label ("rectangle");
  gtk_signal_connect (GTK_OBJECT (button), "clicked", GTK_SIGNAL_FUNC (set_rectangle), NULL);
  gtk_box_pack_start(GTK_BOX(box2), button, FALSE, FALSE, 0);
  gtk_widget_show(button);

  button = gtk_button_new_with_label ("line");
  gtk_signal_connect (GTK_OBJECT (button), "clicked", GTK_SIGNAL_FUNC (set_line), NULL);
  gtk_box_pack_start(GTK_BOX(box2), button, FALSE, FALSE, 0);
  gtk_widget_show(button);

  button = gtk_button_new_with_label ("fill");
  gtk_signal_connect (GTK_OBJECT (button), "clicked", GTK_SIGNAL_FUNC (set_fill), NULL);
  gtk_box_pack_start(GTK_BOX(box2), button, FALSE, FALSE, 0);
  gtk_widget_show(button);

  button = gtk_button_new_with_label ("fuzzy fill");
  gtk_signal_connect (GTK_OBJECT (button), "clicked", GTK_SIGNAL_FUNC (set_fuzzy_fill), NULL);
  gtk_box_pack_start(GTK_BOX(box2), button, FALSE, FALSE, 0);
  gtk_widget_show(button);

  button = gtk_button_new_with_label ("eye dropper");
  gtk_signal_connect (GTK_OBJECT (button), "clicked", GTK_SIGNAL_FUNC (set_sample), NULL);
  gtk_box_pack_start(GTK_BOX(box2), button, FALSE, FALSE, 0);
  gtk_widget_show(button);

  button = gtk_button_new_with_label ("zoom");
  gtk_signal_connect (GTK_OBJECT (button), "clicked", GTK_SIGNAL_FUNC (set_zoom), NULL);
  gtk_box_pack_start(GTK_BOX(box2), button, FALSE, FALSE, 0);
  gtk_widget_show(button);

  tool_label = gtk_label_new("brush");
  gtk_box_pack_start(GTK_BOX(box2), tool_label, TRUE, TRUE, 0);
  gtk_widget_show(tool_label);
  set_brush();

  coords_label = gtk_label_new("X:    0 Y:    0 ");
  gtk_box_pack_start(GTK_BOX(box2), coords_label, FALSE, FALSE, 0);
  gtk_widget_show(coords_label);

  separator = gtk_hseparator_new();
  gtk_box_pack_start(GTK_BOX(window_box), separator, TRUE, TRUE, 0);
  gtk_widget_show(separator);

  /* table */
  drawing_table = gtk_table_new(5,3,FALSE);
  gtk_box_pack_start(GTK_BOX(window_box), drawing_table, TRUE, TRUE, 0);

  /* (m) */
  text = gtk_label_new("meters");
  gtk_table_attach(GTK_TABLE(drawing_table), text,
		   0,2,0,1,
		   GTK_SHRINK|GTK_FILL,
		   GTK_SHRINK|GTK_FILL,
                   1,1);
  text = gtk_label_new("");
  gtk_table_attach(GTK_TABLE(drawing_table), text,
		   4,5,0,3,
		   GTK_EXPAND|GTK_SHRINK|GTK_FILL,
		   GTK_EXPAND|GTK_SHRINK|GTK_FILL,
                   1,1);

  text = gtk_label_new("");
  gtk_table_attach(GTK_TABLE(drawing_table), text,
		   0,1,1,3,
		   GTK_EXPAND|GTK_SHRINK|GTK_FILL,
		   GTK_EXPAND|GTK_SHRINK|GTK_FILL,
                   1,1);

  /* scale */

  hrule = gtk_hruler_new();
  vrule = gtk_vruler_new();
  gtk_ruler_set_metric( GTK_RULER(hrule), GTK_PIXELS);
  gtk_ruler_set_metric( GTK_RULER(vrule), GTK_PIXELS);

  gtk_table_attach(GTK_TABLE(drawing_table), hrule,
		   2,3,0,1,
		   GTK_SHRINK|GTK_FILL,
		   GTK_FILL,
                   1,1);
  gtk_table_attach(GTK_TABLE(drawing_table), vrule,
		   1,2,1,2,
		   GTK_FILL,
		   GTK_SHRINK|GTK_FILL,
                   1,1);

  /* create the scroll bars for the drawing area */
  x_adjustment = gtk_adjustment_new(xstart, 0, 1, 1, 
				    1, 1);

  gtk_signal_connect (GTK_OBJECT (x_adjustment), "value_changed",
		      GTK_SIGNAL_FUNC (horizontal), NULL);
  gtk_signal_connect (GTK_OBJECT (x_adjustment), "changed",
		      GTK_SIGNAL_FUNC (horizontal), NULL);
  hscroll = gtk_hscrollbar_new(GTK_ADJUSTMENT(x_adjustment));
  gtk_table_attach(GTK_TABLE(drawing_table), hscroll, 2,3,2,3,
		   GTK_FILL, 0, 1, 1);

  y_adjustment = gtk_adjustment_new(1, 0, 0, 1, 1, 1);
  gtk_signal_connect (GTK_OBJECT (y_adjustment), "value_changed",
		      GTK_SIGNAL_FUNC (vertical), NULL);
  gtk_signal_connect (GTK_OBJECT (y_adjustment), "changed",
		      GTK_SIGNAL_FUNC (vertical), NULL);
  vscroll = gtk_vscrollbar_new(GTK_ADJUSTMENT(y_adjustment));
  gtk_table_attach(GTK_TABLE(drawing_table), vscroll, 3,4,1,2,
		   0, GTK_FILL, 1,1);

  /* Create the drawing area */

  drawing_area = gtk_drawing_area_new ();
  gtk_widget_set_events (drawing_area, GDK_EXPOSURE_MASK
			 | GDK_BUTTON_PRESS_MASK
			 | GDK_POINTER_MOTION_MASK
			 | GDK_BUTTON_RELEASE_MASK);

  gtk_table_attach(GTK_TABLE(drawing_table), drawing_area,
		   2,3,1,2, 0, 0, 1,1);
  
  /* Signals used to handle backing pixmap */
  gtk_signal_connect (GTK_OBJECT (drawing_area), "expose_event", 
		      (GtkSignalFunc) expose_event, NULL);
  gtk_signal_connect (GTK_OBJECT (drawing_area), "button_press_event",
		      (GtkSignalFunc) button_press_event, NULL);
  gtk_signal_connect (GTK_OBJECT (drawing_area), "motion_notify_event",
		      (GtkSignalFunc) motion_notify_event, NULL);
  gtk_signal_connect (GTK_OBJECT (drawing_area), "button_release_event",
		      (GtkSignalFunc) button_release_event, NULL); 

  separator = gtk_hseparator_new();
  gtk_box_pack_start(GTK_BOX(window_box), separator, FALSE, FALSE, 0);
  gtk_widget_show(separator);

  /* buttons */
  box2 = gtk_hbox_new(FALSE, 0);
  gtk_box_pack_start(GTK_BOX(window_box), box2, TRUE, TRUE, 0);
  gtk_widget_show(box2);

  separator = gtk_vseparator_new();
  gtk_box_pack_start(GTK_BOX(box2), separator, FALSE, FALSE, 5);
  gtk_widget_show(separator);

  /*ink */
  box3 = gtk_vbox_new(TRUE, 0);
  gtk_box_pack_start(GTK_BOX(box2), box3, TRUE, TRUE, 0);
  gtk_widget_show(box3);

  box4 = gtk_hbox_new(FALSE, 0);
  gtk_box_pack_start(GTK_BOX(box3), box4, TRUE, TRUE, 0);
  gtk_widget_show(box4);
	
  box5 = gtk_vbox_new(TRUE, 0);
  gtk_box_pack_start(GTK_BOX(box4), box5, TRUE, TRUE, 0);
  gtk_widget_show(box5);

  button = gtk_button_new_with_label ("unknown");
  gtk_signal_connect (GTK_OBJECT (button), "clicked", 
		      GTK_SIGNAL_FUNC (set_unknown), NULL);
  gtk_box_pack_start(GTK_BOX(box5), button, FALSE, TRUE, 0);
  gtk_widget_show(button);

  button = gtk_button_new_with_label ("offlimits");
  gtk_signal_connect (GTK_OBJECT (button), "clicked", 
		      GTK_SIGNAL_FUNC (set_offlimits), NULL);
  gtk_box_pack_start(GTK_BOX(box5), button, FALSE, TRUE, 0);
  gtk_widget_show(button);

  if(ink < -1) {
    ink = .90;
    color = carmen_graphics_add_color_rgb(255.0*(1.0-ink),
					  255.0*(1.0-ink), 
					  255.0*(1.0-ink));
  }

  adjustment = gtk_adjustment_new(ink, 0.0, 1.0, 0.01, 0.1, 0.0);
  gtk_signal_connect (GTK_OBJECT (adjustment), "value_changed",
		      GTK_SIGNAL_FUNC (set_ink), NULL);
  gtk_signal_connect (GTK_OBJECT (adjustment), "changed",
		      GTK_SIGNAL_FUNC (set_ink), NULL);
  ink_scale = gtk_hscale_new(GTK_ADJUSTMENT(adjustment));
  gtk_scale_set_draw_value(GTK_SCALE(ink_scale), FALSE);

  box5 = gtk_vbox_new(TRUE, 0);
  gtk_box_pack_start(GTK_BOX(box4), box5, TRUE, TRUE, 0);
  gtk_widget_show(box5);
  
  text = gtk_label_new("Ink (Probability)");
  gtk_box_pack_start(GTK_BOX(box5), text, TRUE, TRUE, 0);
  gtk_widget_show(text);
  
  gtk_box_pack_start(GTK_BOX(box5), ink_scale, TRUE, TRUE, 0);
  gtk_widget_show(ink_scale);
  
  sprintf(ink_text, "%.2f", ink);
  ink_label = gtk_label_new(ink_text);
  gtk_box_pack_start(GTK_BOX(box5), ink_label, TRUE, TRUE, 0);
  gtk_widget_show(ink_label);

  separator = gtk_vseparator_new();
  gtk_box_pack_start(GTK_BOX(box2), separator, FALSE, FALSE, 5);
  gtk_widget_show(separator);

  /* fuzzyness */
  box3 = gtk_vbox_new(TRUE, 0);
  gtk_box_pack_start(GTK_BOX(box2), box3, TRUE, TRUE, 0);
  gtk_widget_show(box3);

  text = gtk_label_new("Fuzziness");
  gtk_box_pack_start(GTK_BOX(box3), text, TRUE, TRUE, 0);
  gtk_widget_show(text);

  adjustment = gtk_adjustment_new(fuzzyness, 0.0, 1.0, 0.01, 0.1, 0.0);
  gtk_signal_connect (GTK_OBJECT (adjustment), "value_changed",
		      GTK_SIGNAL_FUNC (set_fuzzyness), NULL);
  fuzzy_scale = gtk_hscale_new(GTK_ADJUSTMENT(adjustment));
  gtk_scale_set_draw_value(GTK_SCALE(fuzzy_scale), FALSE);

  gtk_box_pack_start(GTK_BOX(box3), fuzzy_scale, TRUE, TRUE, 0);
  gtk_widget_show(fuzzy_scale);

  sprintf(fuzzyness_text, "%.2f", fuzzyness);
  fuzzyness_label = gtk_label_new(fuzzyness_text);
  gtk_box_pack_start(GTK_BOX(box3), fuzzyness_label, TRUE, TRUE, 0);
  gtk_widget_show(fuzzyness_label);

  separator = gtk_vseparator_new();
  gtk_box_pack_start(GTK_BOX(box2), separator, FALSE, FALSE, 5);
  gtk_widget_show(separator);

  /* brush size */
  box3 = gtk_vbox_new(TRUE, 0);
  gtk_box_pack_start(GTK_BOX(box2), box3, TRUE, TRUE, 0);
  gtk_widget_show(box3);

  text = gtk_label_new("Brush size");
  gtk_box_pack_start(GTK_BOX(box3), text, TRUE, TRUE, 0);
  gtk_widget_show(text);

  box4 = gtk_hbox_new(FALSE, 0);
  gtk_box_pack_start(GTK_BOX(box3), box4, TRUE, TRUE, 0);
  gtk_widget_show(box4);

  /* brush decr */
  button = gtk_button_new_with_label ("smaller");
  gtk_signal_connect (GTK_OBJECT (button), "clicked", GTK_SIGNAL_FUNC (brush_decr), NULL);
  gtk_box_pack_start(GTK_BOX(box4), button, TRUE, TRUE, 0);
  gtk_widget_show(button);

  /* brush incr */
  button = gtk_button_new_with_label ("larger");
  gtk_signal_connect (GTK_OBJECT (button), "clicked", GTK_SIGNAL_FUNC (brush_incr), NULL);
  gtk_box_pack_start(GTK_BOX(box4), button, TRUE, TRUE, 0);
  gtk_widget_show(button);

  sprintf(brush_text, "%d", carmen_trunc(2.0 * brush_size));
  brush_label = gtk_label_new(brush_text);
  gtk_box_pack_start(GTK_BOX(box3), brush_label, FALSE, FALSE, 0);
  gtk_widget_show(brush_label);

  separator = gtk_vseparator_new();
  gtk_box_pack_start(GTK_BOX(box2), separator, FALSE, FALSE, 5);
  gtk_widget_show(separator);

  /* line_size */
  box3 = gtk_vbox_new(TRUE, 0);
  gtk_box_pack_start(GTK_BOX(box2), box3, TRUE, TRUE, 0);
  gtk_widget_show(box3);

  text = gtk_label_new("Line size");
  gtk_box_pack_start(GTK_BOX(box3), text, TRUE, TRUE, 0);
  gtk_widget_show(text);

  box4 = gtk_hbox_new(FALSE, 0);
  gtk_box_pack_start(GTK_BOX(box3), box4, TRUE, TRUE, 0);
  gtk_widget_show(box4);

  button = gtk_button_new_with_label ("smaller");
  gtk_signal_connect (GTK_OBJECT (button), "clicked", GTK_SIGNAL_FUNC (line_decr), NULL);
  gtk_box_pack_start(GTK_BOX(box4), button, TRUE, TRUE, 0);
  gtk_widget_show(button);

  button = gtk_button_new_with_label ("larger");
  gtk_signal_connect (GTK_OBJECT (button), "clicked", GTK_SIGNAL_FUNC (line_incr), NULL);
  gtk_box_pack_start(GTK_BOX(box4), button, TRUE, TRUE, 0);
  gtk_widget_show(button);

  sprintf(line_text, "%d", carmen_trunc(line_size));
  line_label = gtk_label_new(line_text);
  gtk_box_pack_start(GTK_BOX(box3), line_label, FALSE, FALSE, 0);
  gtk_widget_show(line_label);
  
  separator = gtk_vseparator_new();
  gtk_box_pack_start(GTK_BOX(box2), separator, FALSE, FALSE, 5);
  gtk_widget_show(separator);

  /* filled */
  box3 = gtk_vbox_new(TRUE, 0);
  gtk_box_pack_start(GTK_BOX(box2), box3, TRUE, TRUE, 0);
  gtk_widget_show(box3);

  text = gtk_label_new("Shape fill");
  gtk_box_pack_start(GTK_BOX(box3), text, TRUE, TRUE, 0);
  gtk_widget_show(text);

  box4 = gtk_hbox_new(FALSE, 0);
  gtk_box_pack_start(GTK_BOX(box3), box4, TRUE, TRUE, 0);
  gtk_widget_show(box4);

  button = gtk_button_new_with_label ("not filled");
  gtk_signal_connect (GTK_OBJECT (button), "clicked", 
		      GTK_SIGNAL_FUNC (set_not_filled), NULL);
  gtk_box_pack_start(GTK_BOX(box4), button, TRUE, TRUE, 0);
  gtk_widget_show(button);

  button = gtk_button_new_with_label ("filled");
  gtk_signal_connect (GTK_OBJECT (button), "clicked", 
		      GTK_SIGNAL_FUNC (set_filled), NULL);
  gtk_box_pack_start(GTK_BOX(box4), button, TRUE, TRUE, 0);
  gtk_widget_show(button);

  fill_label = gtk_label_new("not filled");
  gtk_box_pack_start(GTK_BOX(box3), fill_label, FALSE, FALSE, 0);
  gtk_widget_show(fill_label);

  separator = gtk_vseparator_new();
  gtk_box_pack_start(GTK_BOX(box2), separator, FALSE, FALSE, 5);
  gtk_widget_show(separator);

  if (map)
    set_up_map_widgets();
  else
    gtk_widget_show(window);

  place_font = gdk_font_load ("fixed");

}

/* calls the necessary graphics initialization functions */
void init_graphics(int argc, char **argv)
{
  gnome_init("map_editor", "1.0", argc, argv);
  gdk_imlib_init();	
  carmen_graphics_setup_colors();

}
