/*********************************************************
 *
 * This source code is part of the Carnegie Mellon Robot
 * Navigation Toolkit (CARMEN)
 *
 * CARMEN Copyright (c) 2002 Michael Montemerlo, Nicholas
 * Roy, and Sebastian Thrun
 *
 * CARMEN is free software; you can redistribute it and/or 
 * modify it under the terms of the GNU General Public 
 * License as published by the Free Software Foundation; 
 * either version 2 of the License, or (at your option)
 * any later version.
 *
 * CARMEN is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied 
 * warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
 * PURPOSE.  See the GNU General Public License for more 
 * details.
 *
 * You should have received a copy of the GNU General 
 * Public License along with CARMEN; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place, 
 * Suite 330, Boston, MA  02111-1307 USA
 *
 ********************************************************/

/**********************************************
 * library of function for mapserver clients  *
 **********************************************/
#include <carmen/carmen.h>
#ifndef NO_ZLIB
#include <zlib.h>
#endif

carmen_map_t **map_update;
carmen_handler_t *map_update_handler_external;

/* subscribe to incoming gridmap messages */

static int context_array_size = 0;
static IPC_CONTEXT_PTR *context_array = NULL;

static int
get_context_id(void)
{
  int index = 0;
  IPC_CONTEXT_PTR current_context = IPC_getContext();

  if (context_array == NULL)
    return -1;

  while (context_array[index] != NULL) 
    {
      if (context_array[index] == current_context)
	return index;
      index++;
    }

  return -1;

}

static int 
add_context(void)
{
  int index;
  
  if (context_array == NULL)
    {
      context_array = (IPC_CONTEXT_PTR *)calloc(10, sizeof(IPC_CONTEXT_PTR));
      carmen_test_alloc(context_array);

      map_update = (carmen_map_t **)
	calloc(10, sizeof(carmen_map_t *));
      carmen_test_alloc(map_update);

      map_update_handler_external = (carmen_handler_t *)
	calloc(10, sizeof(carmen_handler_t));
      carmen_test_alloc(map_update_handler_external);

      context_array_size = 10;
      context_array[0] = IPC_getContext();

      return 0;
    }

  index = 0;
  while (index < context_array_size && context_array[index] != NULL) 
    index++;
  
  if (index == context_array_size) 
    {
      context_array_size += 10;
      context_array = (IPC_CONTEXT_PTR *)realloc
	(context_array, context_array_size*sizeof(IPC_CONTEXT_PTR));
      carmen_test_alloc(context_array);
      memset(context_array+index, 0, 10*sizeof(IPC_CONTEXT_PTR));
    }

  context_array[index] = IPC_getContext();
  return index;
}

static void 
map_update_interface_handler(MSG_INSTANCE msgRef, BYTE_ARRAY callData,
			     void *clientData __attribute__ ((unused)))
{
  IPC_RETURN_TYPE err = IPC_OK;
  FORMATTER_PTR formatter;
  int context_id;
  carmen_grid_map_message map_msg;
  carmen_map_t *new_map;
  int i;

#ifndef NO_ZLIB
  int uncompress_return;
  int uncompress_size;
  int uncompress_size_result;
#endif

  context_id = get_context_id();

  if (context_id < 0) 
    {
      carmen_warn("Bug detected: invalid context\n");
      IPC_freeByteArray(callData);
      return;
    }

  formatter = IPC_msgInstanceFormatter(msgRef);
  err = IPC_unmarshallData(formatter, callData, &map_msg, 
			   sizeof(carmen_grid_map_message));
  IPC_freeByteArray(callData);

  carmen_test_ipc_return(err, "Could not unmarshall", 
			 IPC_msgInstanceName(msgRef));

  new_map = map_update[context_id];
  if (new_map->map != NULL)
    free(new_map->map);
  if (new_map->complete_map != NULL)
    free(new_map->complete_map);
  
  new_map->config = map_msg.config;
  
  new_map->complete_map = (float *)
    calloc(new_map->config.x_size*new_map->config.y_size, sizeof(float));
  carmen_test_alloc(new_map->complete_map);
  
#ifndef NO_ZLIB
  if (map_msg.compressed) 
    {
      uncompress_size = new_map->config.x_size*new_map->config.y_size*
	sizeof(float);
      uncompress_size_result = uncompress_size;
      uncompress_return = uncompress((void *)new_map->complete_map,   
				     (uLong *)&uncompress_size_result,
				     (void *)map_msg.map, 
				     map_msg.size);
    } 
  else 
    {
      memcpy(new_map->complete_map, map_msg.map, map_msg.size);
    }
#else
  if (map_msg.compressed) 
    {
      carmen_warn("Received compressed map from server. This program was\n"
		  "compiled without zlib support, so this map cannot be\n"
		  "used. Sorry.\n");
      
      memset(new_map, 0, sizeof(carmen_map_t));
      
      if (map_msg.map)
	free(map_msg.map);
      if (map_msg.err_mesg)
	free(map_msg.err_mesg);
      
      return;
    } 
  else 
    {
      memcpy(new_map->complete_map, map_msg.map, map_msg.size);
    }
#endif 
  new_map->map = (float **)
    calloc(map_msg.config.x_size, sizeof(float *));
  carmen_test_alloc(new_map->map);
  for(i = 0; i < map_msg.config.x_size; i++)
    new_map->map[i] = new_map->complete_map + 
      i*map_msg.config.y_size;

  if (map_msg.map)
    free(map_msg.map);
  if (map_msg.err_mesg)
    free(map_msg.err_mesg);

  if (map_update_handler_external[context_id])
    map_update_handler_external[context_id](map_update[context_id]);
}

void 
carmen_map_subscribe_gridmap_update_message(carmen_map_t *map,
					    carmen_handler_t handler,
					    carmen_subscribe_t subscribe_how) 
{
  IPC_RETURN_TYPE err = IPC_OK;  
  int context_id;

  err = IPC_defineMsg(CARMEN_MAP_GRIDMAP_UPDATE_NAME, IPC_VARIABLE_LENGTH, 
		      CARMEN_MAP_GRIDMAP_UPDATE_FMT);
  carmen_test_ipc_exit(err, "Could not define message", 
		       CARMEN_MAP_GRIDMAP_UPDATE_NAME);

  if (subscribe_how == CARMEN_UNSUBSCRIBE) 
    {
      IPC_unsubscribe(CARMEN_MAP_GRIDMAP_UPDATE_NAME, 
		      map_update_interface_handler);
      return;
    }

  context_id = get_context_id();
  if (context_id < 0)
    context_id = add_context();

  if (map)
    map_update[context_id] = map;
  else if (map_update[context_id] == NULL)
    {
      map_update[context_id] = 
	(carmen_map_t *)calloc
	(1, sizeof(carmen_map_t));
      carmen_test_alloc(map_update[context_id]);
    }

  map_update_handler_external[context_id] = handler;

  err = IPC_subscribe(CARMEN_MAP_GRIDMAP_UPDATE_NAME, 
		      map_update_interface_handler, NULL);
  if (subscribe_how == CARMEN_SUBSCRIBE_LATEST)
    IPC_setMsgQueueLength(CARMEN_MAP_GRIDMAP_UPDATE_NAME, 1);
  else
    IPC_setMsgQueueLength(CARMEN_MAP_GRIDMAP_UPDATE_NAME, 100);

  carmen_test_ipc(err, "Could not subscribe", CARMEN_MAP_GRIDMAP_UPDATE_NAME);
}

/* send a request for a gridmap */
int
carmen_map_get_gridmap(carmen_map_p client_map)
{
  IPC_RETURN_TYPE err;
  static carmen_gridmap_request query;
  static carmen_grid_map_message *response;
  unsigned int timeout = 10000;
  int i;
#ifndef NO_ZLIB
  int uncompress_return;
  int uncompress_size;
  int uncompress_size_result;
#endif

  err = IPC_defineMsg(CARMEN_GRIDMAP_REQUEST_NAME, IPC_VARIABLE_LENGTH, 
		      CARMEN_GRIDMAP_REQUEST_FMT);
  carmen_test_ipc_exit(err, "Could not define message", 
		       CARMEN_GRIDMAP_REQUEST_NAME);

  strcpy(query.host, carmen_get_tenchar_host_name());
  query.timestamp = carmen_get_time_ms();
  err = IPC_queryResponseData(CARMEN_GRIDMAP_REQUEST_NAME, &query, 
			      (void **)&response, timeout);
  if (err != IPC_OK)
    {
      carmen_test_ipc(err, "Could not get map_request", 
		      CARMEN_GRIDMAP_REQUEST_NAME);
      carmen_warn("\nDid you remember to start the mapserver, or give a map "
		  "to the paramServer?\n");
      return -1;
    }

  if (response->size == 0) 
    {
      carmen_warn("Error receiving map: %s\n", response->err_mesg);
      return -1;
    }

  if (client_map)
    {
      client_map->config = response->config;

      client_map->complete_map = (float *)
	calloc(client_map->config.x_size*client_map->config.y_size, 
	       sizeof(float));
      carmen_test_alloc(client_map->complete_map);

#ifndef NO_ZLIB
      if (response->compressed) 
	{
	  uncompress_size = client_map->config.x_size*
	    client_map->config.y_size*sizeof(float);
	  uncompress_size_result = uncompress_size;
	  uncompress_return = uncompress
	    ((void *)client_map->complete_map,   
	     (uLong *)&uncompress_size_result,
	     (void *)response->map, response->size);
	} 
      else 
	{
	  memcpy(client_map->complete_map, response->map, response->size);
	}
#else
      if (response->compressed) 
	{
	  carmen_warn("Received compressed map from server. This program was\n"
		      "compiled without zlib support, so this map cannot be\n"
		      "used. Sorry.\n");
	  
	  memset(client_map, 0, sizeof(carmen_map_t));
	  
	  if (response->map)
	    free(response->map);
	  if (response->err_mesg)
	    free(response->err_mesg);
	  free(response);
	  
	  return -1;
	} 
      else 
	{
	  memcpy(client_map->complete_map, response->map, response->size);
	}
#endif 
      client_map->map = (float **)
	calloc(response->config.x_size, sizeof(float *));
      carmen_test_alloc(client_map->map);
      for(i = 0; i < response->config.x_size; i++)
	client_map->map[i] = client_map->complete_map + 
	  i*response->config.y_size;	      
    }
  
  if (response->map)
    free(response->map);
  if (response->err_mesg)
    free(response->err_mesg);
  free(response);

  return 0;
}

/*
void 
carmen_placelist_interface_handler(MSG_INSTANCE msgRef, BYTE_ARRAY callData, 
				   void *clientData __attribute__ ((unused)))
{
  carmen_map_placelist_message internal_placelist;
  FORMATTER_PTR formatter;
  IPC_RETURN_TYPE err = IPC_OK;
  int i;

  formatter = IPC_msgInstanceFormatter(msgRef);

  if (client_placelist) 
    err = IPC_unmarshallData(formatter, callData, &internal_placelist, 
			     sizeof(carmen_map_placelist_message));

  IPC_freeByteArray(callData);

  carmen_test_ipc_return(err, "Could not unmarshall", 
			 IPC_msgInstanceName(msgRef));
  
  if (client_placelist)
    {
      client_placelist->num_places = internal_placelist.num_places;
       client_placelist->places = (carmen_place_p)calloc
	(internal_placelist.num_places, sizeof(carmen_place_t));
      carmen_test_alloc(client_placelist->places);
      for (i = 0; i < client_placelist->num_places; i++)
	memcpy(&client_placelist->places[i],&internal_placelist.places[i],
	       sizeof(carmen_place_t));
      free (internal_placelist.places);
    }
  if (internal_placelist_handler)
    internal_placelist_handler();
}
*/

/* subscribe to incoming gridmap messages */
void 
carmen_map_subscribe_placelist_message(carmen_map_placelist_p placelist 
				       __attribute__ ((unused)), 
				       void (*handler)(void) 
				       __attribute__ ((unused)))
{
  carmen_warn("Subscribing to placelist messages is currently unsupported.\n");

  return;
}

/* send a request for a gridmap */
int
carmen_map_get_placelist(carmen_map_placelist_p placelist)
{

  IPC_RETURN_TYPE err;
  static carmen_map_placelist_request req;
  static carmen_map_placelist_message *response;
  unsigned int timeout = 10000;
	
  err = IPC_defineMsg(CARMEN_PLACELIST_REQUEST_NAME, IPC_VARIABLE_LENGTH, 
		      CARMEN_PLACELIST_REQUEST_FMT);
  carmen_test_ipc_exit(err, "Could not define message", 
		       CARMEN_PLACELIST_REQUEST_NAME);

  strcpy(req.host, carmen_get_tenchar_host_name());
  req.timestamp = carmen_get_time_ms();

  err = IPC_queryResponseData(CARMEN_PLACELIST_REQUEST_NAME, &req, 
			      (void **)&response, timeout);
  carmen_test_ipc_return_int(err, "Could not get placelist", 
			     CARMEN_PLACELIST_REQUEST_NAME);

  if (placelist) 
    {
      placelist->places = response->places;
      placelist->num_places = response->num_places;
    }
  else 
    free(response->places);

  free(response);

  return 0;
}

/* send a request for a gridmap */
int
carmen_map_get_offlimits(carmen_offlimits_p *offlimits, int *list_length)
{

  IPC_RETURN_TYPE err;
  static carmen_map_offlimits_request query;
  static carmen_map_offlimits_message *response;
  unsigned int timeout = 10000;

  err = IPC_defineMsg(CARMEN_OFFLIMITS_REQUEST_NAME, IPC_VARIABLE_LENGTH, 
		      CARMEN_OFFLIMITS_REQUEST_FMT);
  carmen_test_ipc_exit(err, "Could not define message", 
		       CARMEN_OFFLIMITS_REQUEST_NAME);

  strcpy(query.host, carmen_get_tenchar_host_name());
  query.timestamp = carmen_get_time_ms();

  err = IPC_queryResponseData(CARMEN_OFFLIMITS_REQUEST_NAME, &query, 
			      (void **)&response, timeout);
  if (err != IPC_OK)
    {
      carmen_test_ipc(err, "Could not get map_request", 
		      CARMEN_OFFLIMITS_REQUEST_NAME);
      carmen_warn("\nDid you remember to start the mapserver, or give "
		  "a map to the paramServer?\n");
      return -1;
    }

  if (offlimits)
    {
      *list_length = response->list_length;
      *offlimits = response->offlimits_list;
    }

  free(response);

  return 0;
}

int 
carmen_map_apply_offlimits_chunk_to_map(carmen_offlimits_p offlimits_list,
					int list_length, carmen_map_p map)
{
  int current_x, current_y, x, y;
  carmen_bresenham_param_t b_params;
  int i;
  carmen_offlimits_t cur_seg;

  /* This function requires a map to markup */
  if(offlimits_list == NULL || list_length == 0 ||
     map == NULL || map->complete_map == NULL)
    return -1;

  for(i = 0; i < list_length; i++) 
    {
      cur_seg = offlimits_list[i];
      switch(cur_seg.type) 
	{
	case CARMEN_OFFLIMITS_POINT_ID:
	  if(cur_seg.x1 >= 0 && cur_seg.y1 >= 0 && 
	     cur_seg.x1 < map->config.x_size && 
	     cur_seg.y1 < map->config.y_size)
	    map->map[cur_seg.x1][cur_seg.y1] += 2.0;
	  break;
	  
	case CARMEN_OFFLIMITS_LINE_ID:
	  carmen_get_bresenham_parameters(cur_seg.x1, cur_seg.y1, cur_seg.x2, 
					  cur_seg.y2, &b_params);
	  do 
	    {
	      carmen_get_current_point(&b_params, &current_x, &current_y);
	      if(current_x >= 0 && current_y >= 0 &&
		 current_x < map->config.x_size && 
		 current_y < map->config.y_size)
		map->map[current_x][current_y] += 2.0;
	    } 
	  while(carmen_get_next_point(&b_params));
	  break;
	case CARMEN_OFFLIMITS_RECT_ID:
	  for(x = cur_seg.x1; x <= cur_seg.x2; x++)
	    for(y = cur_seg.y1; y <= cur_seg.y2; y++)
	      if(x >= 0 && y >= 0 && 
		 x < map->config.x_size && y < map->config.y_size)
		map->map[x][y] += 2.0;
	  break;
	default:
	  break;
	}
    }
  return 0;
}

inline int 
carmen_map_to_world(carmen_map_point_p carmen_map_point, 
		    carmen_world_point_p world_point) 
{
  double x = carmen_map_point->x * carmen_map_point->map->config.resolution;
  double y = carmen_map_point->y * carmen_map_point->map->config.resolution;
  
  world_point->pose.x = x;
  world_point->pose.y = y;
  world_point->pose.theta = 0;
  world_point->map = carmen_map_point->map;

  return 0;
}

inline int 
carmen_world_to_map(carmen_world_point_p world_point, 
		    carmen_map_point_p carmen_map_point) 
{
  int x = 
    carmen_round(world_point->pose.x / world_point->map->config.resolution);
  int y = 
    carmen_round(world_point->pose.y / world_point->map->config.resolution);

  carmen_map_point->x = x;
  carmen_map_point->y = y;

  carmen_map_point->map = world_point->map;

  return 0;
}

inline int 
carmen_map_to_trajectory(carmen_map_point_p carmen_map_point, 
			 carmen_traj_point_p traj_point) 
{
  double x = carmen_map_point->x * carmen_map_point->map->config.resolution;
  double y = carmen_map_point->y * carmen_map_point->map->config.resolution;
  
  traj_point->x = x;
  traj_point->y = y;
  traj_point->theta = 0.0;
  traj_point->t_vel = 0.0;
  traj_point->r_vel = 0.0;

  return 0;
}

inline int
carmen_point_to_map(carmen_point_p point, carmen_map_point_p map_point, 
		    carmen_map_p map)
{
  int x = carmen_round(point->x / map->config.resolution);
  int y = carmen_round(point->y / map->config.resolution);

  map_point->x = x;
  map_point->y = y;
  map_point->map = map;

  return 0;
}

inline int 
carmen_trajectory_to_map(carmen_traj_point_p traj_point, 
			 carmen_map_point_p carmen_map_point,
			 carmen_map_p map) 
{
  int x = carmen_round(traj_point->x / map->config.resolution);
  int y = carmen_round(traj_point->y / map->config.resolution);

  carmen_map_point->x = x;
  carmen_map_point->y = y;

  carmen_map_point->map = map;

  return 0;
}

/* distance between two carmen_map_points in grid cells */
inline double 
carmen_distance_map(carmen_map_point_p p1, carmen_map_point_p p2) 
{
  return hypot(p1->x-p2->x, p1->y-p2->y);
}

inline double 
carmen_distance_world(carmen_world_point_p p1, carmen_world_point_p p2) 
{
  return hypot(p1->pose.x-p2->pose.x, p1->pose.y-p2->pose.y);
}

inline int
carmen_map_cmp(carmen_map_point_p p1, carmen_map_point_p p2)
{
  if (p1->x == p2->x && p1->y == p2->y)
    return 0;

  return 1;
}

inline int
carmen_world_cmp(carmen_world_point_p p1, carmen_world_point_p p2)
{
  if (fabs(p1->pose.x - p2->pose.x) < 1.0 &&
      fabs(p1->pose.y - p2->pose.y) < 1.0 && 
      fabs(carmen_normalize_theta(p1->pose.theta - p2->pose.theta)) < .017)
    return 0;

  return 1;
}

inline carmen_map_p
carmen_map_copy(carmen_map_p map) 
{
  carmen_map_p new_map;
  int i;

  new_map=(carmen_map_p)calloc(1, sizeof(carmen_map_t));
  carmen_test_alloc(new_map);

  *new_map = *map;
  new_map->complete_map=(float *)
    calloc(map->config.x_size*map->config.y_size, sizeof(float));
  carmen_test_alloc(new_map->complete_map);
  memcpy(new_map->complete_map, map->complete_map, 
	 sizeof(float)*map->config.x_size*map->config.y_size);

  new_map->map=(float **)calloc(map->config.x_size, sizeof(float *));
  carmen_test_alloc(new_map->map);

  for (i = 0; i < new_map->config.x_size; i++)
    new_map->map[i] = new_map->complete_map+i*new_map->config.y_size;
  
  return new_map;
}

inline void
carmen_map_destroy(carmen_map_p *map)
{
  free((*map)->complete_map);
  free((*map)->map);
  free((*map));
  *map = NULL;
}

