/*********************************************************
 *
 * This source code is part of the Carnegie Mellon Robot
 * Navigation Toolkit (CARMEN)
 *
 * CARMEN Copyright (c) 2002 Michael Montemerlo, Nicholas
 * Roy, and Sebastian Thrun
 *
 * CARMEN is free software; you can redistribute it and/or 
 * modify it under the terms of the GNU General Public 
 * License as published by the Free Software Foundation; 
 * either version 2 of the License, or (at your option)
 * any later version.
 *
 * CARMEN is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied 
 * warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
 * PURPOSE.  See the GNU General Public License for more 
 * details.
 *
 * You should have received a copy of the GNU General 
 * Public License along with CARMEN; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place, 
 * Suite 330, Boston, MA  02111-1307 USA
 *
 ********************************************************/

#include <carmen/carmen.h>
#include "map_io.h"
#ifndef NO_ZLIB
#include <zlib.h>
#endif

/*****************************************
 * ipc library for the map server        *
 *****************************************/

static char *filename;

/* handler for map requests.
   loads a map from a file and sends it 
   back to the client */

#ifndef NO_ZLIB

static void
compress_map_msg(carmen_grid_map_message *map_msg, carmen_map_t *map)
{
  uLong compress_buf_size;
  int compress_return = -1;

  compress_buf_size = map->config.x_size*map->config.y_size*1.01+12;
  compress_buf_size *= sizeof(float);
  map_msg->map = (unsigned char *)
    calloc(compress_buf_size, sizeof(unsigned char));
  carmen_test_alloc(map_msg->map);
  compress_return = compress((void *)map_msg->map, 
			     (uLong *)&compress_buf_size,
			     (void *)map->complete_map, 
			     (uLong)map->config.x_size * 
			     map->config.y_size*sizeof(float));
  if (compress_return != Z_OK) 
    {
      free(map_msg->map);
      map_msg->map = (unsigned char *)map->complete_map;
      map_msg->size = map->config.x_size * map->config.y_size*sizeof(float);
      map_msg->compressed = 0;
    } 
  else 
    {
      map_msg->size = compress_buf_size;
      map_msg->compressed = 1;
      free(map->complete_map);
    }
}
#endif

static void
assemble_map_msg(carmen_grid_map_message *map_msg)
{
  carmen_map_t map;
  int ret_val;

  ret_val = carmen_map_read_gridmap_chunk(filename, &map);
  if(ret_val < 0)
    {
      memset(map_msg, 0, sizeof(carmen_grid_map_message));
      map_msg->err_mesg = (char *)calloc(17, sizeof(char));
      carmen_test_alloc(map_msg->err_mesg);
      strcpy(map_msg->err_mesg, "Unknown error!");

      map_msg->timestamp = carmen_get_time_ms();
      strcpy(map_msg->host, carmen_get_tenchar_host_name());

      return;
    } 

  free(map.map);

#ifndef NO_ZLIB
  compress_map_msg(map_msg, &map);
#else
  map_msg->map = (unsigned char *)map.complete_map;
  map_msg->size = map.config.x_size * map.config.y_size*sizeof(float);
  map_msg->compressed = 0;
#endif

  map_msg->config = map.config;
  map_msg->err_mesg = (char *)calloc(1, sizeof(char));
  carmen_test_alloc(map_msg->err_mesg);
  map_msg->err_mesg[0] = '\0';

  map_msg->timestamp = carmen_get_time_ms();
  strcpy(map_msg->host, carmen_get_tenchar_host_name());
}

static void 
gridmap_request_handler(MSG_INSTANCE msgRef, BYTE_ARRAY callData, 
			void *clientData __attribute__ ((unused)))
{
  carmen_gridmap_request req;
  carmen_grid_map_message map_msg;
  FORMATTER_PTR formatter;
  IPC_RETURN_TYPE err;

  formatter = IPC_msgInstanceFormatter(msgRef);
  err = IPC_unmarshallData(formatter, callData, &req, 
			   sizeof(carmen_gridmap_request));
  IPC_freeByteArray(callData);  

  carmen_test_ipc_return(err, "Could not unmarshall data", 
			 IPC_msgInstanceName(msgRef));

  assemble_map_msg(&map_msg);

  err = IPC_respondData(msgRef, CARMEN_MAP_GRIDMAP_NAME, &map_msg);  
  carmen_test_ipc(err, "Could not respond", CARMEN_MAP_GRIDMAP_NAME);

  free(map_msg.map);
  free(map_msg.err_mesg);
}


static void 
placelist_request_handler(MSG_INSTANCE msgRef, BYTE_ARRAY callData, 
			  void *clientData __attribute__ ((unused)))
{
  carmen_map_placelist_request req;
  carmen_map_placelist_t placelist;
  carmen_map_placelist_message placelist_msg;
  FORMATTER_PTR formatter;
  IPC_RETURN_TYPE err;

  formatter = IPC_msgInstanceFormatter(msgRef);
  err = IPC_unmarshallData(formatter, callData, &req, 
			   sizeof(carmen_map_placelist_request));
  IPC_freeByteArray(callData);  

  carmen_test_ipc_return(err, "Could not unmarshall data", 
			 IPC_msgInstanceName(msgRef));

  if (carmen_map_chunk_exists(filename, CARMEN_MAP_PLACES_CHUNK) > 0 && 
      carmen_map_read_places_chunk(filename, &placelist) == 0) 
    {
      placelist_msg.places = placelist.places;
      placelist_msg.num_places = placelist.num_places;
    }
  else
    {
      placelist_msg.places = NULL;
      placelist_msg.num_places = 0;
      placelist.num_places = 0;
    } 

  placelist_msg.timestamp = carmen_get_time_ms();
  strcpy(placelist_msg.host, carmen_get_tenchar_host_name());
  
  err = IPC_respondData(msgRef, CARMEN_MAP_PLACELIST_NAME, &placelist_msg);  
  carmen_test_ipc(err, "Could not respond", CARMEN_MAP_PLACELIST_NAME);

  if (placelist.num_places > 0)    
    free(placelist.places);
}

static void 
offlimits_request_handler(MSG_INSTANCE msgRef, BYTE_ARRAY callData, 
			  void *clientData __attribute__ ((unused)))
{
  carmen_map_offlimits_request req;
  carmen_offlimits_p offlimits;
  int list_length;
  carmen_map_offlimits_message offlimits_msg;
  FORMATTER_PTR formatter;
  IPC_RETURN_TYPE err;

  formatter = IPC_msgInstanceFormatter(msgRef);
  err = IPC_unmarshallData(formatter, callData, &req, 
			   sizeof(carmen_map_offlimits_request));
  IPC_freeByteArray(callData);  

  carmen_test_ipc_return(err, "Could not unmarshall data", 
			 IPC_msgInstanceName(msgRef));
  
  if(carmen_map_read_offlimits_chunk(filename, &offlimits, &list_length) < 0)
    {
      offlimits_msg.offlimits_list = NULL;
      offlimits_msg.list_length = 0;
    } else {
      offlimits_msg.offlimits_list = offlimits;
      offlimits_msg.list_length = list_length;
    }

  offlimits_msg.timestamp = carmen_get_time_ms();
  strcpy(offlimits_msg.host, carmen_get_tenchar_host_name());
  
  err = IPC_respondData(msgRef, CARMEN_MAP_OFFLIMITS_NAME, &offlimits_msg);  
  carmen_test_ipc(err, "Could not respond", CARMEN_MAP_OFFLIMITS_NAME);

  if (list_length > 0)    
    free(offlimits);
}

/* set up the ipc connection with central
   subscribe to map requests
   define map messages */
int 
carmen_map_initialize_ipc(void)
{
  IPC_RETURN_TYPE err;

  /* define messages created by this module */
  err = IPC_defineMsg(CARMEN_MAP_GRIDMAP_NAME, IPC_VARIABLE_LENGTH, 
		      CARMEN_MAP_GRIDMAP_FMT);
  carmen_test_ipc_exit(err, "Could not define", CARMEN_MAP_GRIDMAP_NAME);

  err = IPC_defineMsg(CARMEN_MAP_GRIDMAP_UPDATE_NAME, IPC_VARIABLE_LENGTH, 
		      CARMEN_MAP_GRIDMAP_UPDATE_FMT);
  carmen_test_ipc_exit(err, "Could not define", 
		       CARMEN_MAP_GRIDMAP_UPDATE_NAME);

  err = IPC_defineMsg(CARMEN_GRIDMAP_REQUEST_NAME, IPC_VARIABLE_LENGTH, 
		      CARMEN_GRIDMAP_REQUEST_FMT);
  carmen_test_ipc_exit(err, "Could not define", CARMEN_GRIDMAP_REQUEST_NAME);

  err = IPC_defineMsg(CARMEN_PLACELIST_REQUEST_NAME, IPC_VARIABLE_LENGTH, 
		      CARMEN_PLACELIST_REQUEST_FMT);
  carmen_test_ipc_exit(err, "Could not define", CARMEN_PLACELIST_REQUEST_NAME);

  err = IPC_defineMsg(CARMEN_MAP_PLACELIST_NAME, IPC_VARIABLE_LENGTH, 
		      CARMEN_MAP_PLACELIST_FMT);
  carmen_test_ipc_exit(err, "Could not define", CARMEN_MAP_PLACELIST_NAME);

  err = IPC_defineMsg(CARMEN_OFFLIMITS_REQUEST_NAME, IPC_VARIABLE_LENGTH, 
		      CARMEN_OFFLIMITS_REQUEST_FMT);
  carmen_test_ipc_exit(err, "Could not define", CARMEN_OFFLIMITS_REQUEST_NAME);

  err = IPC_defineMsg(CARMEN_MAP_OFFLIMITS_NAME, IPC_VARIABLE_LENGTH, 
		      CARMEN_MAP_OFFLIMITS_FMT);
  carmen_test_ipc_exit(err, "Could not define", CARMEN_MAP_OFFLIMITS_NAME);

  /* setup incoming message handlers */
  err = IPC_subscribe(CARMEN_GRIDMAP_REQUEST_NAME, gridmap_request_handler, 
		      NULL);
  carmen_test_ipc(err, "Could not subscribe", CARMEN_GRIDMAP_REQUEST_NAME);
  IPC_setMsgQueueLength(CARMEN_GRIDMAP_REQUEST_NAME, 100);

  err = IPC_subscribe(CARMEN_PLACELIST_REQUEST_NAME, 
		      placelist_request_handler, NULL);
  carmen_test_ipc(err, "Could not subscribe", CARMEN_PLACELIST_REQUEST_NAME);
  IPC_setMsgQueueLength(CARMEN_PLACELIST_REQUEST_NAME, 100);

  err = IPC_subscribe(CARMEN_OFFLIMITS_REQUEST_NAME, 
		      offlimits_request_handler, NULL);
  carmen_test_ipc(err, "Could not subscribe", CARMEN_OFFLIMITS_REQUEST_NAME);
  IPC_setMsgQueueLength(CARMEN_OFFLIMITS_REQUEST_NAME, 100);

  return 0;
}

void
carmen_map_publish_update(void)
{
  carmen_grid_map_message map_msg;
  IPC_RETURN_TYPE err;

  assemble_map_msg(&map_msg);

  err = IPC_publishData(CARMEN_MAP_GRIDMAP_UPDATE_NAME, &map_msg);
  carmen_test_ipc(err, "Could not publish", CARMEN_MAP_GRIDMAP_UPDATE_NAME);

  free(map_msg.map);
  free(map_msg.err_mesg);
}

void
carmen_map_set_filename(char *new_filename)
{
  if (filename != NULL)
    free(filename);

  filename = (char *)calloc(strlen(new_filename)+1, sizeof(char));
  carmen_test_alloc(filename);

  strcpy(filename, new_filename);
  carmen_map_publish_update();
}

