/*********************************************************
 *
 * This source code is part of the Carnegie Mellon Robot
 * Navigation Toolkit (CARMEN)
 *
 * CARMEN Copyright (c) 2002 Michael Montemerlo, Nicholas
 * Roy, and Sebastian Thrun
 *
 * CARMEN is free software; you can redistribute it and/or 
 * modify it under the terms of the GNU General Public 
 * License as published by the Free Software Foundation; 
 * either version 2 of the License, or (at your option)
 * any later version.
 *
 * CARMEN is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied 
 * warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
 * PURPOSE.  See the GNU General Public License for more 
 * details.
 *
 * You should have received a copy of the GNU General 
 * Public License along with CARMEN; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place, 
 * Suite 330, Boston, MA  02111-1307 USA
 *
 ********************************************************/

#include <carmen/carmen.h>
#include "map_io.h"
#include <assert.h>

#ifndef NO_GRAPHICS
#include <gdk_imlib.h>
#endif

int 
carmen_map_read_comment_chunk(carmen_map_file_p fp)
{
  char comment_str[100], c;
  char *err, id[1024];

  do {
    c = carmen_map_fgetc(fp);
    if(c == EOF)
      return -1;
    else if(c == '#') 
      {
	err = carmen_map_fgets(comment_str, 100, fp);
	if(err == NULL)
	  return -1;
      }
    else
      carmen_map_fseek(fp, -1, SEEK_CUR);
  } while(c == '#');

  assert (strlen(CARMEN_MAP_LABEL)+strlen(CARMEN_MAP_VERSION) < 1024);

  if(carmen_map_fread(&id, strlen(CARMEN_MAP_LABEL)+
		      strlen(CARMEN_MAP_VERSION), 1, fp) == 0)
    return -1;
  if(strncmp(id, CARMEN_MAP_LABEL, strlen(CARMEN_MAP_LABEL)) != 0)
    return 1;
  if(strncmp(id+strlen(CARMEN_MAP_LABEL), CARMEN_MAP_VERSION, 
	     strlen(CARMEN_MAP_VERSION)) != 0)
    return 1;

  return 0;
}

int carmen_map_file(char *filename)
{
  carmen_map_file_p fp;
  int comment_return;

  if(!carmen_file_exists(filename))
    return 0;
  fp = carmen_map_fopen(filename, "r");
  if (fp == NULL)
    return 0;

  comment_return = carmen_map_read_comment_chunk(fp);
  carmen_map_fclose(fp);
	
  if (!comment_return)
    return 1;
		
  return 0;
}

int carmen_map_copy_comments(carmen_map_file_p fp_in, carmen_map_file_p fp_out)
{
  char comment_str[100], id[1024];
  int c;
  char *err;

  do {
    c = carmen_map_fgetc(fp_in);
    if(c == EOF)
      return -1;
    else if(c == '#') {
      err = carmen_map_fgets(comment_str, 100, fp_in);
      if(err == NULL)
	return -1;
      else {
	carmen_map_fputc(c, fp_out);
	carmen_map_fprintf(fp_out, "%s", comment_str);
      }
    }
    else
      carmen_map_fseek(fp_in, -1, SEEK_CUR);
  } while(c == '#');
  if(carmen_map_fread(id, strlen(CARMEN_MAP_LABEL)+strlen(CARMEN_MAP_VERSION),
		      1, fp_in) < 1)
    return -1;
  carmen_map_fwrite(id, strlen(CARMEN_MAP_LABEL)+strlen(CARMEN_MAP_VERSION), 
		    1, fp_out);
  return 0;
}

int carmen_map_strip(carmen_map_file_p fp_in, carmen_map_file_p fp_out, 
		     int chunk)
{
  int chunk_id = CARMEN_MAP_NO_CHUNK;
  int chunk_type, chunk_size, done = 0;
  char *buffer;
  
  if (chunk & CARMEN_MAP_PLACES_CHUNK)
    chunk_id |= CARMEN_MAP_PLACES_CHUNK;
  if (chunk & CARMEN_MAP_GRIDMAP_CHUNK)
    chunk_id |= CARMEN_MAP_GRIDMAP_CHUNK;
  if (chunk & CARMEN_MAP_OFFLIMITS_CHUNK)
    chunk_id |= CARMEN_MAP_OFFLIMITS_CHUNK;
  if (chunk & CARMEN_MAP_EXPECTED_CHUNK)
    chunk_id |= CARMEN_MAP_EXPECTED_CHUNK;
  if (chunk & CARMEN_MAP_LASERSCANS_CHUNK)
    chunk_id |= CARMEN_MAP_LASERSCANS_CHUNK;

  if(chunk_id == CARMEN_MAP_NO_CHUNK) 
    {
      carmen_warn("Error: no valid chunkname was found.\n");
      return -1;
    }

  if(carmen_map_copy_comments(fp_in, fp_out) < 0)
    return -1;

  do {
    chunk_type = carmen_map_fgetc(fp_in);
    if(chunk_type == EOF)
      done = 1;
    
    if(!done)
      if(carmen_map_fread(&chunk_size, sizeof(int), 1, fp_in) < 1)
        done = 1;

    if(!done && (chunk_type & chunk_id) == 0) {
      carmen_map_fputc(chunk_type, fp_out);
      carmen_map_fwrite(&chunk_size, sizeof(int), 1, fp_out);
      buffer = (char *)calloc(chunk_size, sizeof(char));
      carmen_test_alloc(buffer);
      if(carmen_map_fread(buffer, chunk_size, 1, fp_in) < 1) {
	fprintf(stderr, "Error: error writing new file: %s\n", 
		strerror(errno));
	return -1;
      }
      carmen_map_fwrite(buffer, chunk_size, 1, fp_out);
      free(buffer);
    }
    else if(!done)
      if(carmen_map_fseek(fp_in, chunk_size, SEEK_CUR) < 0) {
	fprintf(stderr, "Error: error copying map.\n");
	return -1;
      }
		
    if(!done && carmen_map_feof(fp_in))
      done = 1;
  } while(!done);
  return 0;
}

void strip_trailing_spaces(char *str, int len)
{
  int i;
  
  i = len - 1;
  while(i >= 0 && str[i] == ' ')
    i--;
  i++;
  if(i == len)
    i--;
  str[i] = '\0';
}

int carmen_map_advance_to_chunk(carmen_map_file_p fp, int specific_chunk)
{
  int chunk_type, chunk_size, done = 0;

  carmen_map_fseek(fp, 0, SEEK_SET);
  if(carmen_map_read_comment_chunk(fp) < 0) {
    fprintf(stderr, "Error: Could not read comment chunk.\n");
    return -1;
  }
  do {
    chunk_type = carmen_map_fgetc(fp);
    if(chunk_type == EOF)
      done = 1;
    if(chunk_type == specific_chunk) {
      carmen_map_fseek(fp, -1, SEEK_CUR);
      return 0;
    }
    if(!done)
      if(carmen_map_fread(&chunk_size, sizeof(int), 1, fp) < 1)
	done = 1;
    if(!done)
      if(carmen_map_fseek(fp, chunk_size, SEEK_CUR) < 0)
	done = 1;
    if(!done && carmen_map_feof(fp))
      done = 1;
  } while(!done);
  return -1;
}

int carmen_map_chunk_exists(char *filename, int specific_chunk)
{
  carmen_map_file_p fp;
  char chunk_type;
  int chunk_size;

  fp = carmen_map_fopen(filename, "r");
  if(fp == NULL) {
    fprintf(stderr, "Error: could not open file %s for reading.\n",
	    filename);
    return -1;
  }
  if(carmen_map_advance_to_chunk(fp, specific_chunk) < 0) {
    carmen_map_fclose(fp);
    return 0;
  }
  else {
    chunk_type = carmen_map_fgetc(fp);
    carmen_map_fread(&chunk_size, sizeof(int), 1, fp);
    carmen_map_fclose(fp);
    return chunk_size;
  }
}

int carmen_map_read_creator_chunk(char *filename, time_t *creation_time, 
				  char *username, char *origin, 
				  char *description)
{
  carmen_map_file_p fp;
  int chunk_type, chunk_size;
  char chunk_description[12];

  fp = carmen_map_fopen(filename, "r");
  if(fp == NULL) {
    fprintf(stderr, "Error: could not open file %s for reading.\n",
	    filename);
    return -1;
  }
  if(carmen_map_advance_to_chunk(fp, CARMEN_MAP_CREATOR_CHUNK) < 0) 
    {
      if (carmen_map_advance_to_chunk(fp, 0) < 0) 
	{
	  carmen_warn("You have an old-style map. The creator chunk id "
		      "is wrong.\n\nFIX IT!\n\n");
	}
      else {
	carmen_warn("Error: Could not find a creator chunk.\n");
	carmen_map_fclose(fp);
	return -1;
      }
    }
  chunk_type = carmen_map_fgetc(fp);
  carmen_map_fread(&chunk_size, sizeof(int), 1, fp);
  carmen_map_fread(chunk_description, 10, 1, fp);
  carmen_map_fread(username, 10, 1, fp);
  strip_trailing_spaces(username, 10);
  carmen_map_fread(creation_time, sizeof(time_t), 1, fp);
  carmen_map_fread(origin, 80, 1, fp);
  strip_trailing_spaces(origin, 80);
  carmen_map_fread(description, 80, 1, fp);
  strip_trailing_spaces(description, 80);
  carmen_map_fclose(fp);
  return 0;
}

int 
carmen_map_read_gridmap_config(char *filename, carmen_map_config_p config) 
{
  carmen_map_file_p fp;
  int chunk_type, chunk_size;
  char chunk_description[12];
  int size_x, size_y;
  float resolution;

  if (config == NULL) 
    {
      fprintf(stderr, "Error: config argument is NULL in %s.\n",
	      __FUNCTION__);
      return -1;
    }

  fp = carmen_map_fopen(filename, "r");
  if(fp == NULL) {
    fprintf(stderr, "Error: could not open file %s for reading.\n",
	    filename);
    return -1;
  }

  if(carmen_map_advance_to_chunk(fp, CARMEN_MAP_GRIDMAP_CHUNK) < 0) {
    fprintf(stderr, "Error: Could not find a gridmap chunk.\n");
    fprintf(stderr, "       This file is probably not a map file.\n");
    carmen_map_fclose(fp);
    return -1;
  }
  chunk_type = carmen_map_fgetc(fp);
  carmen_map_fread(&chunk_size, sizeof(int), 1, fp);
  carmen_map_fread(chunk_description, 10, 1, fp);
  carmen_map_fread(&size_x, sizeof(int), 1, fp);
  carmen_map_fread(&size_y, sizeof(int), 1, fp);
  carmen_map_fread(&resolution, sizeof(float), 1, fp);
  
  config->x_size = size_x;
  config->y_size = size_y;
  config->resolution = resolution;
  config->map_name = (char *)calloc(strlen(filename)+1, sizeof(char));
  carmen_test_alloc(config->map_name);
  strcpy(config->map_name, filename);

  return 0;
}

int carmen_map_read_gridmap_chunk(char *filename, carmen_map_p map)
{
  carmen_map_file_p fp;
  int chunk_type, chunk_size;
  char chunk_description[12];
  int size_x, size_y;
  float resolution;
  int n;

  fp = carmen_map_fopen(filename, "r");
  if(fp == NULL) {
    fprintf(stderr, "Error: could not open file %s for reading.\n",
		       filename);
    return -1;
  }
  if(carmen_map_advance_to_chunk(fp, CARMEN_MAP_GRIDMAP_CHUNK) < 0) {
    fprintf(stderr, "Error: Could not find a gridmap chunk.\n");
    fprintf(stderr, "       This file is probably not a map file.\n");
    carmen_map_fclose(fp);
    return -1;
  }
  chunk_type = carmen_map_fgetc(fp);
  carmen_map_fread(&chunk_size, sizeof(int), 1, fp);
  carmen_map_fread(chunk_description, 10, 1, fp);
  carmen_map_fread(&size_x, sizeof(int), 1, fp);
  carmen_map_fread(&size_y, sizeof(int), 1, fp);
  carmen_map_fread(&resolution, sizeof(float), 1, fp);
  
  map->config.x_size = size_x;
  map->config.y_size = size_y;
  map->config.resolution = resolution;
  map->config.map_name = (char *)calloc(strlen(filename)+1, sizeof(char));
  carmen_test_alloc(map->config.map_name);
  strcpy(map->config.map_name, filename);

  map->complete_map = 
    (float *)calloc(map->config.x_size * map->config.y_size, sizeof(float));
  carmen_test_alloc(map->complete_map);
  map->map = (float **)calloc(map->config.x_size, sizeof(float *));
  carmen_test_alloc(map->map);
  for(n = 0; n < map->config.x_size; n++)
    map->map[n] = map->complete_map + n * map->config.y_size;
  carmen_map_fread(map->complete_map, sizeof(float) * size_x * size_y, 1, fp);

  carmen_map_fclose(fp);
  return 0;
}

int carmen_map_read_offlimits_chunk(char *filename, 
				    carmen_offlimits_p *offlimits_list,
				    int *list_length)
{
  carmen_map_file_p fp;
  char chunk_type, chunk_description[12];
  int chunk_size, num_points, num_lines, num_rects, i;
  carmen_offlimits_t cur_seg;

  if (offlimits_list == NULL || list_length == NULL)
    return -1;

  *list_length = 0;
  
  fp = carmen_map_fopen(filename, "r");
  if(fp == NULL) {
    fprintf(stderr, "Error: could not open file %s for reading.\n",
	    filename);
    return -1;
  }
  if(carmen_map_advance_to_chunk(fp, CARMEN_MAP_OFFLIMITS_CHUNK) < 0) {
    carmen_map_fclose(fp);
    return -1;
  }
  chunk_type = carmen_map_fgetc(fp);
  carmen_map_fread(&chunk_size, sizeof(int), 1, fp);
  carmen_map_fread(chunk_description, 10, 1, fp);
  carmen_map_fread(&num_points, sizeof(int), 1, fp);
  if (num_points > 0) {
    *list_length = num_points;
    *offlimits_list = (carmen_offlimits_p)calloc
      (num_points, sizeof(carmen_offlimits_t));
    carmen_test_alloc(*offlimits_list);
    for(i = 0; i < num_points; i++) {
      cur_seg.type = CARMEN_OFFLIMITS_POINT_ID;
      carmen_map_fread(&(cur_seg.x1), sizeof(int), 1, fp);
      carmen_map_fread(&(cur_seg.y1), sizeof(int), 1, fp);
      cur_seg.x2 = 0;
      cur_seg.y2 = 0;
      (*offlimits_list)[i] = cur_seg;
    }
  }
  carmen_map_fread(&num_lines, sizeof(int), 1, fp);
  if (num_lines > 0)
    {
      if (*list_length == 0)
	{
	  *offlimits_list = (carmen_offlimits_p)calloc
	    (num_lines, sizeof(carmen_offlimits_t));
	  carmen_test_alloc(*offlimits_list);
	  *list_length = num_lines;
	}
      else {
	*list_length += num_lines;
	*offlimits_list = (carmen_offlimits_p)realloc
	  (*offlimits_list, (*list_length)*
	   sizeof(carmen_offlimits_t));
	carmen_test_alloc(*offlimits_list);
      }

      for(i = 0; i < num_lines; i++) {
	cur_seg.type = CARMEN_OFFLIMITS_LINE_ID;
	carmen_map_fread(&(cur_seg.x1), sizeof(int), 1, fp);
	carmen_map_fread(&(cur_seg.y1), sizeof(int), 1, fp);
	carmen_map_fread(&(cur_seg.x2), sizeof(int), 1, fp);
	carmen_map_fread(&(cur_seg.y2), sizeof(int), 1, fp);
	(*offlimits_list)[i+num_points] = cur_seg;
      }
    }
  carmen_map_fread(&num_rects, sizeof(int), 1, fp);
  if (num_rects > 0)
    {
      if (*list_length == 0)				
	{
	  *list_length = num_rects;
	  *offlimits_list = (carmen_offlimits_p)calloc
	    (num_rects, sizeof(carmen_offlimits_t));
	  carmen_test_alloc(*offlimits_list);
	}
      else 
	{
	  *list_length += num_rects;
	  *offlimits_list = (carmen_offlimits_p)realloc
	    (*offlimits_list, (*list_length)*
	     sizeof(carmen_offlimits_t));
	  carmen_test_alloc(*offlimits_list);
	}
      for(i = 0; i < num_rects; i++) {
	cur_seg.type = CARMEN_OFFLIMITS_RECT_ID;
	carmen_map_fread(&(cur_seg.x1), sizeof(int), 1, fp);
	carmen_map_fread(&(cur_seg.y1), sizeof(int), 1, fp);
	carmen_map_fread(&(cur_seg.x2), sizeof(int), 1, fp);
	carmen_map_fread(&(cur_seg.y2), sizeof(int), 1, fp);
	(*offlimits_list)[i+num_points+num_lines] = cur_seg;
      }
    }
  return 0;
}

int carmen_map_read_offlimits_chunk_into_map(char *filename, carmen_map_p map)
{
  carmen_offlimits_p offlimits_list;
  int list_length;
	
  if (carmen_map_read_offlimits_chunk
      (filename, &offlimits_list, &list_length) < 0)
    return -1;

  if (carmen_map_apply_offlimits_chunk_to_map
      (offlimits_list, list_length, map) < 0)
    return -1;

  free(offlimits_list);
  return 0;
}

int carmen_map_read_places_chunk(char *filename, carmen_map_placelist_p places)
{
  carmen_map_file_p fp;
  int chunk_type, chunk_size, i;
  char chunk_description[12];
  int place_type;
  float float_var;

  fp = carmen_map_fopen(filename, "r");
  if(fp == NULL) {
    fprintf(stderr, "Error: could not open file %s for reading.\n",
	    filename);
    return -1;
  }
  if(carmen_map_advance_to_chunk(fp, CARMEN_MAP_PLACES_CHUNK) < 0) {
    fprintf(stderr, "Error: Could not find a places chunk.\n");
    carmen_map_fclose(fp);
    return -1;
  }
  chunk_type = carmen_map_fgetc(fp);
  carmen_map_fread(&chunk_size, sizeof(int), 1, fp);
  carmen_map_fread(chunk_description, 10, 1, fp);
  carmen_map_fread(&(places->num_places), sizeof(int), 1, fp);
  places->places = (carmen_place_p)calloc(places->num_places, 
					  sizeof(carmen_place_t));
  carmen_test_alloc(places->places);
  for(i = 0; i < places->num_places; i++) {
    carmen_map_fread(&place_type, sizeof(int), 1, fp);
    carmen_map_fread(places->places[i].name, 20, 1, fp);
    strip_trailing_spaces(places->places[i].name, 20);
    if(place_type == CARMEN_NAMED_POSITION_TYPE) {
      carmen_map_fread(&float_var, sizeof(float), 1, fp);
      places->places[i].x = float_var;
      carmen_map_fread(&float_var, sizeof(float), 1, fp);
      places->places[i].y = float_var;
    }
    else if(place_type == CARMEN_NAMED_POSE_TYPE) {
      carmen_map_fread(&float_var, sizeof(float), 1, fp);
      places->places[i].x = float_var;
      carmen_map_fread(&float_var, sizeof(float), 1, fp);
      places->places[i].y = float_var;
      carmen_map_fread(&float_var, sizeof(float), 1, fp);
      places->places[i].theta = float_var;
    }
    else if(place_type == CARMEN_LOCALIZATION_INIT_TYPE) {
      carmen_map_fread(&float_var, sizeof(float), 1, fp);
      places->places[i].x = float_var;
      carmen_map_fread(&float_var, sizeof(float), 1, fp);
      places->places[i].y = float_var;
      carmen_map_fread(&float_var, sizeof(float), 1, fp);
      places->places[i].theta = float_var;

      carmen_map_fread(&float_var, sizeof(float), 1, fp);
      places->places[i].x_std = float_var;
      carmen_map_fread(&float_var, sizeof(float), 1, fp);
      places->places[i].y_std = float_var;
      carmen_map_fread(&float_var, sizeof(float), 1, fp);
      places->places[i].theta_std = float_var;
    }
    places->places[i].type = place_type;
  }
  carmen_map_fclose(fp);
  return 0;
}

int carmen_map_read_laserscans_chunk(char *filename, 
				     carmen_laser_scan_p *scan_list, 
				     int *num_scans)
{
  carmen_map_file_p fp;
  int chunk_type, chunk_size, i, j;
  char chunk_description[12];
  float float_var;

  fp = carmen_map_fopen(filename, "r");
  if(fp == NULL) {
    fprintf(stderr, "Error: could not open file %s for reading.\n",
	    filename);
    return -1;
  }
  if(carmen_map_advance_to_chunk(fp, CARMEN_MAP_LASERSCANS_CHUNK) < 0) {
    fprintf(stderr, "Error: Could not find a laserscans chunk.\n");
    carmen_map_fclose(fp);
    return -1;
  }
  chunk_type = carmen_map_fgetc(fp);
  carmen_map_fread(&chunk_size, sizeof(int), 1, fp);
  carmen_map_fread(chunk_description, 10, 1, fp);
  carmen_map_fread(num_scans, sizeof(int), 1, fp);
  *scan_list = 
    (carmen_laser_scan_p)calloc(*num_scans, sizeof(carmen_laser_scan_t));
  carmen_test_alloc(*scan_list);
  for(i = 0; i < *num_scans; i++) {
    carmen_map_fread(&float_var, sizeof(float), 1, fp);
    (*scan_list)[i].x = float_var;
    carmen_map_fread(&float_var, sizeof(float), 1, fp);
    (*scan_list)[i].y = float_var;
    carmen_map_fread(&float_var, sizeof(float), 1, fp);
    (*scan_list)[i].theta = float_var;

    carmen_map_fread(&((*scan_list)[i].num_readings), sizeof(int), 1, fp);

    (*scan_list)[i].range = (float *)calloc((*scan_list)[i].num_readings, 
					    sizeof(float));
    carmen_test_alloc((*scan_list)[i].range);
    for (j = 0; j < (*scan_list)[i].num_readings; j++) {
      carmen_map_fread(&float_var, sizeof(float), 1, fp);
      (*scan_list)[i].range[j] = float_var;
    }
  }
  return 0;
}



/* image to map */

#ifndef NO_GRAPHICS

int carmen_map_image_to_map_color_unknown(unsigned char r, unsigned char g, unsigned char b) {
  if (   ( r == 0 && g == 0 &&   b==255) ||
	 ( r == 0 && g == 255 && b==0) )
    return 1;
  return 0;
}

float carmen_map_image_to_map_color_occupancy(unsigned char r, unsigned char g, unsigned char b) {
  float occ =( ((float)r) + ((float)g) + ((float)b)  ) / (3.0 * 255.0);
  if (occ < 0.001)
    occ = 0.001;
  if (occ > 0.999)
    occ = 0.999;
  return 1.0 - occ;
}

carmen_map_p
carmen_map_image_to_map( GdkImlibImage* im, double resolution )
{
  carmen_map_p map;
  int x_index, y_index;

  if (im == NULL) {
    carmen_warn("Error: im = NULL in %s at line %d in file %s\n",
		__FUNCTION__, __LINE__, __FILE__);
    return NULL;
  }

  map = (carmen_map_p)calloc(1, sizeof(carmen_map_t));
  carmen_test_alloc(map);

  map->config.x_size = im->rgb_width;
  map->config.y_size = im->rgb_height;
  map->config.resolution =  resolution;

  map->complete_map = (float *)
    calloc(map->config.x_size*map->config.y_size, sizeof(float));
  carmen_test_alloc(map->complete_map);
  
  map->map = (float **)calloc(map->config.x_size, sizeof(float *));
  carmen_test_alloc(map->map);
  for (x_index = 0; x_index < map->config.x_size; x_index++) 
    map->map[x_index] = map->complete_map+x_index*map->config.y_size;

  for (x_index = 0; x_index < map->config.x_size; x_index++) 
    for (y_index = 0; y_index < map->config.y_size; y_index++)       {
	unsigned char r,g ,b;
	r = im->rgb_data[3*(x_index + y_index * im->rgb_width) + 0];
	g = im->rgb_data[3*(x_index + y_index * im->rgb_width) + 1];
	b = im->rgb_data[3*(x_index + y_index * im->rgb_width) + 2];

	if (carmen_map_image_to_map_color_unknown(r,g,b)) 
	  map->map[x_index][map->config.y_size-1-y_index] = -1.0;
	else 
	  map->map[x_index][map->config.y_size-1-y_index] =  carmen_map_image_to_map_color_occupancy(r,g,b);
      }
  return map;
}



carmen_map_p
carmen_map_imagefile_to_map(char *filename, double resolution) {
  carmen_map_p map;
  GdkImlibImage* im;

  im = (GdkImlibImage*) gdk_imlib_load_image(filename);
  if (im == NULL)
    carmen_die_syserror("Couldn't open %s for reading", filename);

  map = carmen_map_image_to_map(im, resolution);

  gdk_imlib_kill_image(im);

 return map;
}

#endif


