/*********************************************************
 *
 * This source code is part of the Carnegie Mellon Robot
 * Navigation Toolkit (CARMEN)
 *
 * CARMEN Copyright (c) 2002 Michael Montemerlo, Nicholas
 * Roy, and Sebastian Thrun
 *
 * CARMEN is free software; you can redistribute it and/or 
 * modify it under the terms of the GNU General Public 
 * License as published by the Free Software Foundation; 
 * either version 2 of the License, or (at your option)
 * any later version.
 *
 * CARMEN is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied 
 * warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
 * PURPOSE.  See the GNU General Public License for more 
 * details.
 *
 * You should have received a copy of the GNU General 
 * Public License along with CARMEN; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place, 
 * Suite 330, Boston, MA  02111-1307 USA
 *
 ********************************************************/

#include <carmen/carmen.h>
#include "map_io.h"

void write_fixed_string(carmen_map_file_p fp, char *str, int n)
{
  int l, i;

  if (str == NULL)
    l = 0;
  else
    l = strlen(str);
  
  if(l > n)
    l = n;
  for(i = 0; i < l; i++)
    carmen_map_fprintf(fp, "%c", str[i]);
  if(l < n)
    for(i = l; i < n; i++)
      carmen_map_fprintf(fp, " ");
}

int carmen_map_write_all(carmen_map_file_p fp, float **prob,
			 int size_x, int size_y, double resolution,
			 char *comment_origin, char *comment_description,
			 char *creator_origin, char *creator_description,
			 carmen_place_p places, int num_places,
			 carmen_offlimits_p offlimits_list, int offlimits_num_items,
			 carmen_laser_scan_p scan_list, int num_scans)
{
  if (fp == NULL || prob == NULL)
    return -1;

  if (comment_origin == NULL)
    comment_origin = "";
  if (comment_description == NULL)
    comment_description = "";
  if (carmen_map_write_comment_chunk(fp, size_x, size_y, resolution, comment_origin,
				     comment_description) < 0)
    return -1;

  if (carmen_map_write_id(fp) < 0)
    return -1;

  if (creator_origin == NULL)
    creator_origin = "";
  if (creator_description == NULL)
    creator_description = "";
  if (carmen_map_write_creator_chunk(fp, creator_origin, creator_description) < 0)
    return -1;

  if (carmen_map_write_gridmap_chunk(fp, prob, size_x, size_y, resolution) < 0)
    return -1;

  if (places != NULL && num_places > 0)
    if (carmen_map_write_places_chunk(fp, places, num_places) < 0)
      return -1;

  if (offlimits_list != NULL && offlimits_num_items > 0)
    if (carmen_map_write_offlimits_chunk(fp, offlimits_list, offlimits_num_items) < 0)
      return -1;

  if (scan_list != NULL && num_scans > 0)
    if (carmen_map_write_laserscans_chunk(fp, scan_list, num_scans) < 0)
      return -1;

  return 0;
}

int carmen_map_write_comment_chunk(carmen_map_file_p fp, int size_x, int size_y, 
				 double resolution, char *origin,
				 char *description)
{
  time_t t = time(NULL);

  carmen_map_fprintf(fp, "################################################"
		     "#####\n");
  carmen_map_fprintf(fp, "#\n");
  carmen_map_fprintf(fp, "# Carnegie Mellon Robot Toolkit (CARMEN) "
		     "map file\n");
  carmen_map_fprintf(fp, "#\n");
  if (getlogin() == NULL)
    carmen_map_fprintf(fp, "# Map author    : UNKNOWN\n");
  else
    carmen_map_fprintf(fp, "# Map author    : %s\n", getlogin());
  carmen_map_fprintf(fp, "# Creation date : %s", asctime(localtime(&t)));
  carmen_map_fprintf(fp, "# Map size      : %d x %d\n", size_x, size_y);
  carmen_map_fprintf(fp, "# Resolution    : %.1f\n", resolution);
  carmen_map_fprintf(fp, "# Origin        : %s\n", origin);
  carmen_map_fprintf(fp, "# Description   : %s", description);
  carmen_map_fprintf(fp, "#\n");
  carmen_map_fprintf(fp, "###############################################"
		     "######\n");
  return 0;
}

int carmen_map_write_id(carmen_map_file_p fp)
{
  carmen_map_fprintf(fp, "%s%s", CARMEN_MAP_LABEL, CARMEN_MAP_VERSION);
  return 0;
}

int carmen_map_write_creator_chunk(carmen_map_file_p fp, char *origin, 
				   char *description)
{
  int chunk_size;
  time_t t;

  carmen_map_fputc(CARMEN_MAP_CREATOR_CHUNK, fp);
  chunk_size = 10 + 10 + sizeof(time_t) + 80 + 80;
  carmen_map_fwrite(&chunk_size, sizeof(int), 1, fp);
  carmen_map_fprintf(fp, "CREATOR   ");
  if (getlogin() == NULL)
    write_fixed_string(fp, "UNKNOWN", 10);
  else
    write_fixed_string(fp, getlogin(), 10);
  t = time(NULL);
  carmen_map_fwrite(&t, sizeof(time_t), 1, fp);
  write_fixed_string(fp, origin, 80);
  write_fixed_string(fp, description, 80);
  return 0;
}

int carmen_map_write_gridmap_chunk(carmen_map_file_p fp, float **prob, 
				 int size_x, int size_y, double resolution)
{
  int x, chunk_size;
  float local_resolution = resolution;

  carmen_map_fputc(CARMEN_MAP_GRIDMAP_CHUNK, fp);
  chunk_size = 10 + 12 + size_x * size_y * 4;
  carmen_map_fwrite(&chunk_size, sizeof(int), 1, fp);
  carmen_map_fprintf(fp, "GRIDMAP   ");
  carmen_map_fwrite(&size_x, sizeof(int), 1, fp);
  carmen_map_fwrite(&size_y, sizeof(int), 1, fp);
  carmen_map_fwrite(&local_resolution, sizeof(float), 1, fp);
  for(x = 0; x < size_x; x++)
    carmen_map_fwrite(prob[x], sizeof(float) * size_y, 1, fp);
  return 0;
}

int carmen_map_write_places_chunk(carmen_map_file_p fp, carmen_place_p places, 
				int num_places)
{
  int chunk_size, i;
  float float_var;

  carmen_map_fputc(CARMEN_MAP_PLACES_CHUNK, fp);
  chunk_size = 10 + 4 + 24 * num_places;
  for(i = 0; i < num_places; i++)
    if(places[i].type == CARMEN_NAMED_POSITION_TYPE)
      chunk_size += 8;
    else if(places[i].type == CARMEN_NAMED_POSE_TYPE)
      chunk_size += 12;
    else if(places[i].type == CARMEN_LOCALIZATION_INIT_TYPE)
      chunk_size += 24;
  carmen_map_fwrite(&chunk_size, sizeof(int), 1, fp);
  carmen_map_fprintf(fp, "PLACES    ");
  carmen_map_fwrite(&num_places, sizeof(int), 1, fp);
  for(i = 0; i < num_places; i++) {
    carmen_map_fwrite(&(places[i].type), sizeof(int), 1, fp);
    write_fixed_string(fp, places[i].name, 20);
    if(places[i].type == CARMEN_NAMED_POSITION_TYPE) {
      float_var = places[i].x;
      carmen_map_fwrite(&float_var, sizeof(float), 1, fp);
      float_var = places[i].y;
      carmen_map_fwrite(&float_var, sizeof(float), 1, fp);
    }
    else if(places[i].type == CARMEN_NAMED_POSE_TYPE) {
      float_var = places[i].x;
      carmen_map_fwrite(&float_var, sizeof(float), 1, fp);
      float_var = places[i].y;
      carmen_map_fwrite(&float_var, sizeof(float), 1, fp);
      float_var = places[i].theta;
      carmen_map_fwrite(&float_var, sizeof(float), 1, fp);
    }
    else if(places[i].type == CARMEN_LOCALIZATION_INIT_TYPE) {
      float_var = places[i].x;
      carmen_map_fwrite(&float_var, sizeof(float), 1, fp);
      float_var = places[i].y;
      carmen_map_fwrite(&float_var, sizeof(float), 1, fp);
      float_var = places[i].theta;
      carmen_map_fwrite(&float_var, sizeof(float), 1, fp);
      float_var = places[i].x_std;
      carmen_map_fwrite(&float_var, sizeof(float), 1, fp);
      float_var = places[i].y_std;
      carmen_map_fwrite(&float_var, sizeof(float), 1, fp);
      float_var = places[i].theta_std;
      carmen_map_fwrite(&float_var, sizeof(float), 1, fp);
    }
  }
  return 0;
}

int carmen_map_write_offlimits_chunk(carmen_map_file_p fp, 
				     carmen_offlimits_p offlimits_list,
				     int num_items)
{
  int chunk_size, i;
  int num_points = 0, num_lines = 0, num_rects = 0;

  carmen_map_fputc(CARMEN_MAP_OFFLIMITS_CHUNK, fp);
  chunk_size = 10 + 4 + 4 + 4;
  for(i = 0; i < num_items; i++)
    if(offlimits_list[i].type == CARMEN_OFFLIMITS_POINT_ID)
      num_points++;
    else if(offlimits_list[i].type == CARMEN_OFFLIMITS_LINE_ID)
      num_lines++;
    else if(offlimits_list[i].type == CARMEN_OFFLIMITS_RECT_ID)
      num_rects++;
  chunk_size += num_points * 8 + num_lines * 16 + num_rects * 16;

  carmen_map_fwrite(&chunk_size, sizeof(int), 1, fp);
  carmen_map_fprintf(fp, "OFFLIMITS ");
  carmen_map_fwrite(&num_points, sizeof(int), 1, fp);
  for(i = 0; i < num_items; i++)
    if(offlimits_list[i].type == CARMEN_OFFLIMITS_POINT_ID) {
      carmen_map_fwrite(&offlimits_list[i].x1, sizeof(int), 1, fp);
      carmen_map_fwrite(&offlimits_list[i].y1, sizeof(int), 1, fp);
    }
  carmen_map_fwrite(&num_lines, sizeof(int), 1, fp);
  for(i = 0; i < num_lines; i++)
    if(offlimits_list[i].type == CARMEN_OFFLIMITS_LINE_ID) {
      carmen_map_fwrite(&offlimits_list[i].x1, sizeof(int), 1, fp);
      carmen_map_fwrite(&offlimits_list[i].y1, sizeof(int), 1, fp);
      carmen_map_fwrite(&offlimits_list[i].x2, sizeof(int), 1, fp);
      carmen_map_fwrite(&offlimits_list[i].y2, sizeof(int), 1, fp);
    }
  carmen_map_fwrite(&num_rects, sizeof(int), 1, fp);
  for(i = 0; i < num_items; i++)
    if(offlimits_list[i].type == CARMEN_OFFLIMITS_RECT_ID) {
      carmen_map_fwrite(&offlimits_list[i].x1, sizeof(int), 1, fp);
      carmen_map_fwrite(&offlimits_list[i].y1, sizeof(int), 1, fp);
      carmen_map_fwrite(&offlimits_list[i].x2, sizeof(int), 1, fp);
      carmen_map_fwrite(&offlimits_list[i].y2, sizeof(int), 1, fp);
    }
  return 0;
}

int carmen_map_write_laserscans_chunk(carmen_map_file_p fp, 
				      carmen_laser_scan_p scan_list,
				      int num_scans)
{
  int chunk_size;
  int i, j;
  float float_var;

  carmen_map_fputc(CARMEN_MAP_LASERSCANS_CHUNK, fp);
  chunk_size = 10 + 4 + 16 * num_scans;
  for(i = 0; i < num_scans; i++)
    chunk_size += 4 * scan_list[i].num_readings;
  carmen_map_fwrite(&chunk_size, sizeof(int), 1, fp);
  carmen_map_fprintf(fp, "LASERSCANS");
  carmen_map_fwrite(&num_scans, sizeof(int), 1, fp);
  for(i = 0; i < num_scans; i++) {
    float_var = scan_list[i].x;
    carmen_map_fwrite(&float_var, sizeof(float), 1, fp);
    float_var = scan_list[i].y;
    carmen_map_fwrite(&float_var, sizeof(float), 1, fp);
    float_var = scan_list[i].theta;
    carmen_map_fwrite(&float_var, sizeof(float), 1, fp);
    carmen_map_fwrite(&(scan_list[i].num_readings), sizeof(int), 1, fp);
    for (j = 0; j < scan_list[i].num_readings; j++) {
      float_var = scan_list[i].range[j];
      carmen_map_fwrite(&float_var, sizeof(float), 1, fp);
    }
  }
  return 0;
}
