/*********************************************************
 *
 * This source code is part of the Carnegie Mellon Robot
 * Navigation Toolkit (CARMEN)
 *
 * CARMEN Copyright (c) 2002 Michael Montemerlo, Nicholas
 * Roy, and Sebastian Thrun
 *
 * CARMEN is free software; you can redistribute it and/or 
 * modify it under the terms of the GNU General Public 
 * License as published by the Free Software Foundation; 
 * either version 2 of the License, or (at your option)
 * any later version.
 *
 * CARMEN is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied 
 * warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
 * PURPOSE.  See the GNU General Public License for more 
 * details.
 *
 * You should have received a copy of the GNU General 
 * Public License along with CARMEN; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place, 
 * Suite 330, Boston, MA  02111-1307 USA
 *
 ********************************************************/

#include <carmen/carmen.h>

#include <limits.h>

#define        ERROR                    -1


#define        PIONEER_WATCHDOG_TIMEOUT          2.0
#define        PIONEER_SERIAL_TIMEOUT            5.0

#define        BACKWARDS_FLAG           0

#define        METRES_PER_INCH          0.0254
#define        WHEELBASE                (16.5 * METRES_PER_INCH) // need to set this
#define        ROT_VEL_FACT_RAD         (WHEELBASE)

#ifdef _REENTRANT
#include <pthread.h>
static pthread_t main_thread;
static int thread_is_running = 0;
static int shutdown_in_progress = 0;
#endif

static char *dev_name;
static int dev_fd;
static int num_sonars = 16;
static int use_sonar = 0;
static int pioneer_version = 1;
static char *robot_sonar_offsets = "";
static double  robot_sensor_angle = 0;

static int set_velocity = 0;
static int command_vl, command_vr;
static double last_command = 0;
static int timeout = 0;

static double robot_acceleration = 1.0;
static double robot_deceleration = 1.0;
static double robot_max_t_vel = 1.0;

static double pioneer_relative_wheel_size = 1.0;
static double pioneer_distance_conversion_factor = 1.0;
static double pioneer_velocity_conversion_factor = 1.0;

/*** Some pioneer internal constants.  I have them defined here since the ***/
/*** outside world doesn't need to know about them.                       ***/

#define PIONEER_PACKET_HEADER  "\xFA\xFB"

#define PIONEER_SYNC0           0
#define PIONEER_SYNC1           1
#define PIONEER_SYNC2           2

#define PIONEER_PULSE           0
#define PIONEER_OPEN            1
#define PIONEER_CLOSE           2
#define PIONEER_POLLING         3
#define PIONEER_ENABLE          4
#define PIONEER_SETA            5
#define PIONEER_VEL             11
#define PIONEER_RVEL            21
#define PIONEER_SETRA           23
#define PIONEER_SONAR           28
#define PIONEER_STOP            29
#define PIONEER_VEL2            32

#define PIONEER_NEGATIVE_INT    0x1B
#define PIONEER_STRING          0x2B
#define PIONEER_POSITIVE_INT    0x3B

#define PIONEER_ON              1
#define PIONEER_OFF             0
#define PIONEER_MOTOR_NO_POWER  0x31
#define PIONEER_MOTOR_STOPPED   0x32
#define PIONEER_MOTOR_MOVING    0x33

/* Cyrill, 14.01.2004
   according to the ActivMedia manual the discretization
   is 4mm/sec, not 5mm/sec. So the factor should be 250.0
   and not 200.0 */
//#define PIONEER1_VEL2_CONVERSION_FACTOR             200.0             // m/s => 5mm/sec
#define PIONEER1_VEL2_CONVERSION_FACTOR             250.0             // m/s => 4mm/sec

#define PIONEER2_VEL2_CONVERSION_FACTOR             50.0              // m/s => 2cm/sec
#define PIONEER1_ANGLE_CONVERSION_FACTOR            0.006136          // = 2*PI/1024
#define PIONEER2_ANGLE_CONVERSION_FACTOR            0.001533          // = 2*PI/4096
#define PIONEER1_VELOCITY_CONVERSION_FACTOR_DEFAULT 0.0025332         // ticks => m/s (Pioneer-I after 1996)
#define PIONEER2_VELOCITY_CONVERSION_FACTOR_DEFAULT 0.001             // mm/s => m/s
#define PIONEER_BATTERY_CONVERSION_FACTOR           0.1
#define PIONEER_COMPASS_CONVERSION_FACTOR           2.0
#define PIONEER_RANGE_CONVERSION_FACTOR             .000268
#define PIONEER_ODOMETRY_ROLLOVER                   300.0

#define PIONEER_MAX_NUM_SONARS  16

#define PIONEER1_TICKS_PER_WHEEL_REVOLUTION                  7875.0
#define PIONEER2_TICKS_PER_WHEEL_REVOLUTION                  630.0

/*** When we receive an information packet back from the robot, it is     ***/
/*** initially read into a raw information packet as if to a big char     ***/
/*** buffer.  It is then converted to a regular information packet, which ***/
/*** contains more useful data types.                                     ***/
struct pioneer_raw_information_packet {
  unsigned char motor_status;
  unsigned char x[2];
  unsigned char y[2];
  unsigned char orientation[2];
  unsigned char vl[2];
  unsigned char vr[2];
  unsigned char battery;
  unsigned char bumper[2];
  unsigned char servo_position[2];
  unsigned char pulse_width[2];
  unsigned char compass_heading;
  unsigned char num_sonar_readings;
  struct {
    unsigned char n;
    unsigned char range[2];
  } sonar[3*PIONEER_MAX_NUM_SONARS];
};

typedef enum {PIONEER_NO_POWER, PIONEER_STOPPED, PIONEER_MOVING} 
motor_status_t;

struct pioneer_information_packet {
  motor_status_t motor_status;
  double x;                              /* cm                             */
  double y;                              /* cm                             */
  double orientation;                    /* radians                        */
  double vl;                             /* cm/s                           */
  double vr;                             /* cm/s                           */
  double battery;                        /* volts                          */
  unsigned char left_bumper;
  unsigned char right_bumper;
  float servo_position;                 /* radians                        */
  float pulse_width;                    
  float compass_heading;                /* radians                        */
  unsigned char num_sonar_readings;
  struct {
    int n;                              /* which sonar was read           */
    float range;                        /* cm                             */
  } sonar[3*PIONEER_MAX_NUM_SONARS];
};

static carmen_base_odometry_message odometry;
static carmen_base_sonar_message sonar;
static int lastrx, lastry;
static double lastpx, lastpy;


/*** TABLE OF CONTENTS ***/

/*** Higher-level robot functions ***/

static void  command_robot(void);
static int   initialize_robot();
static int   read_pioneer_parameters(int argc, char **argv);
static void  reset_status(void);
static void  stop_robot(void);
static void  update_status(void);

/*** Lower-level robot functions and helper functions ***/
static int   pioneer_buf_to_checksum(unsigned char *buf);
static int   pioneer_buf_to_int(unsigned char *buf);
static int   pioneer_buf_to_unsigned_int(unsigned char *buf);
static int   pioneer_calculate_checksum(unsigned char *ptr, int length);
static void  pioneer_close(void);
static int   pioneer_convert_information_packet(
						struct pioneer_raw_information_packet *r,
						struct pioneer_information_packet *p);
static void  pioneer_int_to_buf(unsigned char *buf, int n);
static void  pioneer_enable(void);
static void  pioneer_open(void);
static void  pioneer_pulse(void);
static int   pioneer_read_string(unsigned char *buf, double read_timeout);
static int   pioneer_send_command0(char cmd);
static int   pioneer_send_command1(char cmd, int arg1);
static int   pioneer_send_string(unsigned char *ptr, int length);
static void  pioneer_seta(int accel, int decel);
static void  pioneer_setra(int accel, int decel);
static void  pioneer_sonar(int b);
static void  pioneer_stop(void);
static int   pioneer_sync0(void);
static int   pioneer_sync1(void);
static int   pioneer_sync2(void);
static void  pioneer_vel2(int vl, int vr);

static void 
command_robot(void)
{
  double time_elapsed = carmen_get_time_ms() - last_command;
  if(set_velocity) 
    {
      if(command_vl == 0 && command_vr == 0) 
	{
	  fprintf(stderr, "S");
	  stop_robot();
	}
      else 
	{
	  fprintf(stderr, "V");
	  pioneer_vel2(command_vl, command_vr);
	}

      set_velocity = 0;
      timeout = 0;
      last_command = carmen_get_time_ms();
    }
  else if(time_elapsed > PIONEER_WATCHDOG_TIMEOUT && !timeout) 
    {
      fprintf(stderr, "T");
      stop_robot();
      command_vl = 0;
      command_vr = 0;
      timeout = 1;
    }
}


static int 
initialize_robot(void) 
{
  if (pioneer_sync0() < 0)
    return ERROR;

  if (pioneer_sync1() < 0)
    return ERROR;

  if (pioneer_sync2() < 0)
    return ERROR;

  pioneer_open();
  pioneer_pulse();

  pioneer_sonar(use_sonar);
  if (pioneer_version == 2) 
    {
      pioneer_enable();
      pioneer_seta(robot_acceleration, robot_deceleration);
      pioneer_setra(robot_acceleration/2, robot_deceleration/2);
    }

  return 0;
}

static void
parse_sonar_offsets(carmen_point_p sonar_offsets)
{
  char *so,*data, *curr, *next;
  int sonar_num=0, data_size=20;
 
  so=(char*)calloc(strlen(robot_sonar_offsets), sizeof(char));
  carmen_test_alloc(so);
  memcpy(so,robot_sonar_offsets,strlen(robot_sonar_offsets)*sizeof(char));
  data=(char*)calloc(data_size,sizeof(char));
  carmen_test_alloc(data);
  curr=so;
  while(curr!=NULL && sonar_num<num_sonars)
    {
      
      next=strchr(curr, (int)' ');
      memset(data,0,data_size);
      if(next==NULL)
	{
	  carmen_warn("Incorrect number of arguments in offset string. X\n");
	  break;
	}
      else
	{
	  next[0]='\0';
	  data=strncpy(data,curr, strlen(curr)-strlen(next)+1);
	  if(data==NULL)
	    {
	      carmen_warn("Too few arguments in offset string. X\n");
	      break;
	    }
	  sonar_offsets[sonar_num].x=atof(data);
	  curr=&next[1];
	}

      next=strchr(curr, (int)' ');
      memset(data,0,data_size);
      if(next==NULL)
	{
	  carmen_warn("Incorrect number of arguments in offset string. Y\n");
	  break;
	}
      else
	{
	  next[0]='\0';
	  data=strncpy(data,curr, strlen(curr)-strlen(next)+1);
	  if(data==NULL)
	    {
	      carmen_warn("Too few arguments in offset string. Y\n");
	      break;
	    }
	  sonar_offsets[sonar_num].y=atof(data);
	  curr=&next[1];
	}

      next=strchr(curr, (int)' ');
      memset(data, 0, data_size);
      if(next==NULL)
	{
	  data=strcpy(data,curr);
	  data[strlen(curr)]='\0';
	  if(data==NULL)
	    {
	      carmen_warn("Too few arguments in offset string. THETA\n");
	      break;
	    }
	  sonar_offsets[sonar_num].theta=carmen_normalize_theta(atof(data));
	  curr=NULL;
	}
      else
	{
	  next[0]='\0';
	  data=strncpy(data,curr, strlen(curr)-strlen(next)+1);
	  if(data==NULL)
	    {
	      carmen_warn("Too few arguments in offset string. Y\n");
	      break;
	    }
	  sonar_offsets[sonar_num].theta=carmen_normalize_theta(atof(data));
	  curr=&next[3];
	}
      sonar_num++;
    }
  if(sonar_num<(num_sonars-1))
    carmen_warn("Too few arguments in offset string.  END\n");
}

static void 
reset_status(void) {
  odometry.x = 0.0;
  odometry.y = 0.0;
  odometry.theta = 0.0;
  odometry.tv = 0.0;
  odometry.rv = 0.0;
  odometry.timestamp = 0.0;
  
  if (use_sonar) 
    {
      sonar.num_sonars = num_sonars;
      sonar.range = (double *)calloc(num_sonars, sizeof(double));
      carmen_test_alloc(sonar.range);
      sonar.sonar_offsets = (carmen_point_p)calloc(num_sonars, sizeof(carmen_point_t));
      carmen_test_alloc(sonar.sonar_offsets);
      parse_sonar_offsets(sonar.sonar_offsets);
      sonar.sensor_angle=robot_sensor_angle;
      sonar.timestamp = 0.0;
      strncpy(sonar.host, odometry.host, 10);
    }
}


/*** input in cm/s  ***/
static void 
set_wheel_velocities(double vl, double vr)
{
  if(vl > robot_max_t_vel)
    vl = robot_max_t_vel;
  else if(vl < -robot_max_t_vel)
    vl = -robot_max_t_vel;
  if(vr > robot_max_t_vel)
    vr = robot_max_t_vel;
  else if(vr < -robot_max_t_vel)
    vr = -robot_max_t_vel;

  vl /= pioneer_relative_wheel_size;
  vr /= pioneer_relative_wheel_size;

  if(BACKWARDS_FLAG) {
    vl *= -1;
    vr *= -1;
  }

  if (pioneer_version == 1) 
    {
      vl *= PIONEER1_VEL2_CONVERSION_FACTOR;
      vr *= PIONEER1_VEL2_CONVERSION_FACTOR;
    }
  else 
    {
      vl *= PIONEER2_VEL2_CONVERSION_FACTOR;
      vr *= PIONEER2_VEL2_CONVERSION_FACTOR;
    }

  if((command_vl != (int)vl) || (command_vr != (int)vr))
    {
      command_vl = (int)vl;
      command_vr = (int)vr;
      set_velocity = 1;
    }
  last_command = carmen_get_time_ms();
}


static void 
stop_robot(void) 
{
  pioneer_stop();
}


static void 
update_status(void) 
{
  struct pioneer_raw_information_packet r;
  struct pioneer_information_packet p;
  double timestamp;
  int i;
  int no_data = 1;

  do {
    if (pioneer_read_string((char*) &r, PIONEER_SERIAL_TIMEOUT) < 0) 
      {
	carmen_warn("Serial read timed out. Re-sending SYNC\n");
	initialize_robot();
      }
    else
      no_data = 0;
  } while (no_data);

  timestamp = carmen_get_time_ms();

  if (pioneer_convert_information_packet(&r, &p) == ERROR)
    return;

  odometry.x = p.x;
  odometry.y = p.y;
  odometry.theta = p.orientation;
  if (odometry.theta > M_PI)
    odometry.theta -= 2*M_PI;
  odometry.tv = (p.vl + p.vr) / 2.0;
  odometry.rv = (p.vl - p.vr) / ROT_VEL_FACT_RAD;
  odometry.timestamp = timestamp;
  if (use_sonar) 
    {
      for (i = 0; i < p.num_sonar_readings; ++i) 
	sonar.range[p.sonar[i].n] = p.sonar[i].range;
      sonar.timestamp = timestamp;  
    }
}


static int 
pioneer_buf_to_checksum(unsigned char *buf) 
{
  return 256*buf[0] + buf[1];
}


static int 
pioneer_buf_to_int(unsigned char *buf) 
{
  int val = buf[0] + 256*buf[1];
  if (val > SHRT_MAX)
    val -= USHRT_MAX;
  return val;
}


static int 
pioneer_buf_to_unsigned_int(unsigned char *buf) 
{
  int val = buf[0] + 256 * (0x7F & buf[1]);  
  return val;
}


static int 
pioneer_calculate_checksum(unsigned char *ptr, int length) 
{
  unsigned int c = 0;
  
  while(length > 1) 
    {
      c += (*(ptr)<<8) | *(ptr+1);
      c = c & 0xffff;
      length -= 2;
      ptr += 2;
    }

  if(length > 0) 
    c = c ^ (int)*(ptr++);

  return(c);
}


static void 
pioneer_close(void) 
{
  pioneer_send_command0(PIONEER_CLOSE);
}


static int 
pioneer_convert_information_packet (struct pioneer_raw_information_packet *r,
				    struct pioneer_information_packet *p) 
{
  int i;
  int rx, ry, deltax, deltay;
  /* Make sure this is actually a   */
  /* status packet.                 */
  if (r->motor_status != PIONEER_MOTOR_NO_POWER &&
      r->motor_status != PIONEER_MOTOR_STOPPED &&
      r->motor_status != PIONEER_MOTOR_MOVING) 
    {
      carmen_warn("\nTried to convert non-information packet.\n");
      return ERROR;
    }

  rx = pioneer_buf_to_unsigned_int(r->x);
  ry = pioneer_buf_to_unsigned_int(r->y);

  deltax = rx - lastrx;                 /* we must detect and correct for */
  deltay = ry - lastry;                 /* overflow in the odometry       */

  if (deltax > SHRT_MAX/2)
    deltax += SHRT_MIN;
  if (deltax < -SHRT_MAX/2)
    deltax -= SHRT_MIN;
  if (deltay > SHRT_MAX/2)
    deltay += SHRT_MIN;
  if (deltay < -SHRT_MAX/2)
    deltay -= SHRT_MIN;

  p->x = lastpx + (deltax * pioneer_distance_conversion_factor);
  p->y = lastpy + (deltay * pioneer_distance_conversion_factor);

  lastrx = rx; 
  lastry = ry;
  lastpx = p->x;
  lastpy = p->y;
  
  if (pioneer_version == 1)
    p->orientation = pioneer_buf_to_int(r->orientation) * 
      PIONEER1_ANGLE_CONVERSION_FACTOR;
  else
    p->orientation = pioneer_buf_to_int(r->orientation) * 
      PIONEER2_ANGLE_CONVERSION_FACTOR;

  p->vl = pioneer_buf_to_int(r->vl) * pioneer_velocity_conversion_factor * pioneer_relative_wheel_size;
  p->vr = pioneer_buf_to_int(r->vr) * pioneer_velocity_conversion_factor * pioneer_relative_wheel_size;
  p->battery = r->battery * PIONEER_BATTERY_CONVERSION_FACTOR;
  p->left_bumper = r->bumper[1];
  p->right_bumper = r->bumper[0];
  if (pioneer_version == 1)
    p->servo_position = pioneer_buf_to_int(r->servo_position) *
      PIONEER2_ANGLE_CONVERSION_FACTOR;
  else
    p->servo_position = pioneer_buf_to_int(r->servo_position) *
      PIONEER2_ANGLE_CONVERSION_FACTOR;
  p->pulse_width = pioneer_buf_to_unsigned_int(r->pulse_width);
  p->compass_heading = r->compass_heading * PIONEER_COMPASS_CONVERSION_FACTOR;
  if (use_sonar) 
    {
      p->num_sonar_readings = r->num_sonar_readings;
      for (i = 0; i < p->num_sonar_readings; ++i) 
	{
	  p->sonar[i].n = r->sonar[i].n;
	  p->sonar[i].range = pioneer_buf_to_unsigned_int(r->sonar[i].range) *
	    PIONEER_RANGE_CONVERSION_FACTOR;
	}
    }

  return 0;
}


static void 
pioneer_enable(void) 
{
  pioneer_send_command1(PIONEER_ENABLE, PIONEER_ON);
}


static void 
pioneer_int_to_buf(unsigned char *buf, int n) 
{
  if (n >= 0)
    buf[0] = PIONEER_POSITIVE_INT;
  else
    buf[0] = PIONEER_NEGATIVE_INT;

  n = abs(n);

  buf[1] = n & 0x00FF;
  buf[2] = (n & 0xFF00) >> 8;
}
 

static void 
pioneer_open(void) 
{
  pioneer_send_command0(PIONEER_OPEN);
}


static void 
pioneer_pulse(void) 
{
  pioneer_send_command0(PIONEER_PULSE);
}


/*** Return value:  the length of the string read                       ***/
static int 
pioneer_read_string(unsigned char *buf, double read_timeout) 
{
  int c = 0;
  int header_length = strlen(PIONEER_PACKET_HEADER);
  unsigned char header[header_length+1];
  unsigned char length_char[1];
  int i;
  double start_time, current_time;

  for (i = 0; i <= header_length; ++i)
    header[i] = '\0';

  start_time = carmen_get_time_ms();
  do 
    {                                  /* busy waits until we read the   */
      for (i = 1; i < header_length; ++i) /* command header                 */
	header[i-1] = header[i];
      carmen_serial_readn(dev_fd, header + header_length - 1, 1);

      if (read_timeout > 0) {
	current_time = carmen_get_time_ms();
	if (current_time - start_time > read_timeout)
	  // timeout
	  return -1;
      }
    } 
  while (strcmp(header, PIONEER_PACKET_HEADER) != 0);
  
  carmen_serial_readn(dev_fd, length_char, 1);
  c = (int) length_char[0];             /* the length, including chksum   */

  carmen_serial_readn(dev_fd, buf, c);
  c -= 2;                               /* removes checksum from length,  */
                                        /* checksum is now at buf+c       */
  if (pioneer_calculate_checksum(buf, c) != pioneer_buf_to_checksum(buf+c)) 
    {
      fprintf(stderr, "Checksum error on packet ");
      for (i = 0; i < c; ++i)
	fprintf(stderr, "%c", buf[i]);
      fprintf(stderr, "\n");
      return pioneer_read_string(buf, read_timeout);
    }

  return c;
}


static int 
pioneer_send_command0(char cmd) 
{
  char buf[256];

  buf[0] = cmd;
  return pioneer_send_string(buf, 1);
}


static int 
pioneer_send_command1(char cmd, int arg1) 
{
  char buf[256];
  
  buf[0] = cmd;
  pioneer_int_to_buf(buf+1, arg1);
  return pioneer_send_string(buf, 4);
}


static int 
pioneer_send_string(unsigned char *ptr, int length) 
{
  int send_length = length + 5;         /* adds a 2-byte packet header,   */
                                        /* a 1-byte length (inc. chksum), */
                                        /* and a 2-byte checksum          */
  unsigned char cmd[256] = PIONEER_PACKET_HEADER;
  int checksum;
  int i;

  if (length <= 0)
    return 0;

  cmd[2] = (unsigned char) length+2;

  for (i = 0; i < length; ++i)
    cmd[i+3] = ptr[i];

  checksum = pioneer_calculate_checksum(ptr, length);
  cmd[send_length-2] = (unsigned char) ((checksum & 0xFF00) >> 8);
  cmd[send_length-1] = (unsigned char) (checksum & 0x00FF);

  return carmen_serial_writen(dev_fd, cmd, send_length);
}


static void 
pioneer_seta(int accel, int decel) 
{
  accel = abs(accel);
  decel = abs(decel);
  if (pioneer_version == 2)
    {
      accel *= 10;
      decel *= 10;
    }
  pioneer_send_command1(PIONEER_SETA, accel);
  pioneer_send_command1(PIONEER_SETA, -decel);
}


static void 
pioneer_setra(int accel, int decel) 
{
  accel = abs(accel);
  decel = abs(decel);
  if (pioneer_version == 2)
    {
      accel *= 10;
      decel *= 10;
    }
  pioneer_send_command1(PIONEER_SETRA, accel);
  pioneer_send_command1(PIONEER_SETRA, -decel);
}


static void 
pioneer_sonar(int b) 
{
  if (pioneer_version == 1 && b == 0)
    pioneer_send_command1(PIONEER_POLLING, 0);
  else 
    pioneer_send_command1(PIONEER_SONAR, b);
}


static void 
pioneer_stop(void) 
{
  if (pioneer_version == 1)
    pioneer_vel2(0, 0);
  else
    pioneer_send_command0(PIONEER_STOP);
}

  
static int 
pioneer_sync0(void) 
{
  unsigned char buf[256];
  buf[0] = !PIONEER_SYNC0;

  pioneer_send_command0(PIONEER_SYNC0);
  if ((pioneer_read_string(buf, 0.2) == -1) && (pioneer_version == 1)) {
    /* sometimes the Pioneer I's want two syncs */
    pioneer_send_command0(PIONEER_SYNC0);
    if (pioneer_read_string(buf, 0.2) == -1) 
      {
	carmen_warn("Could not SYNC0\n");
	return -1;
      }
  }

  if (buf[0] != PIONEER_SYNC0) {
    carmen_warn("Could not SYNC0\n");
    return -1;
  }

  return 0;
}


static int 
pioneer_sync1(void) 
{
  unsigned char buf[256];

  pioneer_send_command0(PIONEER_SYNC1);
  pioneer_read_string(buf, PIONEER_SERIAL_TIMEOUT);
  if (buf[0] != PIONEER_SYNC1) {
    fprintf(stderr, "Could not SYNC1\n");
    return -1;
  }

  return 0;
}


static int 
pioneer_sync2(void) 
{
  unsigned char buf[256];

  pioneer_send_command0(PIONEER_SYNC2);
  pioneer_read_string(buf, PIONEER_SERIAL_TIMEOUT);

  return 0;
}


static void 
pioneer_vel2(int vl, int vr) 
{
  int arg;
  fprintf(stderr,"+");
  arg = 256*vl + vr;
  pioneer_send_command1(PIONEER_VEL2, arg);
}


static void 
publish_odometry_message(carmen_base_odometry_message *odometry_msg)
{
  IPC_RETURN_TYPE err;
  char *host;
  static int first = 1;

  if(first) {
    host = carmen_get_tenchar_host_name();
    strcpy(odometry_msg->host, host);
    first = 0;
  }

  err = IPC_publishData(CARMEN_BASE_ODOMETRY_NAME, odometry_msg);
  carmen_test_ipc_exit(err, "Could not publish", CARMEN_BASE_ODOMETRY_NAME);
}

static void 
publish_sonar_message(carmen_base_sonar_message *sonar_msg)
{
  IPC_RETURN_TYPE err;
  char *host;
  static int first = 1;

  if(first) {
    host = carmen_get_tenchar_host_name();
    strcpy(sonar_msg->host, host);
    first = 0;
  }

  err = IPC_publishData(CARMEN_BASE_SONAR_NAME, sonar_msg);
  carmen_test_ipc_exit(err, "Could not publish", CARMEN_BASE_SONAR_NAME);
}

static void 
velocity_handler(MSG_INSTANCE msgRef, BYTE_ARRAY callData,
		 void *clientData __attribute__ ((unused)))
{
  IPC_RETURN_TYPE err;
  FORMATTER_PTR formatter;

  carmen_base_velocity_message vel;
  double vl, vr;
  
  formatter = IPC_msgInstanceFormatter(msgRef);
  err = IPC_unmarshallData(formatter, callData, &vel,
			   sizeof(carmen_base_velocity_message));
  IPC_freeByteArray(callData);
  carmen_test_ipc_return(err, "Could not unmarshall", IPC_msgInstanceName(msgRef));

  vl = vel.tv;
  vr = vel.tv;
  vl -= 0.5 * vel.rv * ROT_VEL_FACT_RAD;
  vr += 0.5 * vel.rv * ROT_VEL_FACT_RAD;
  set_wheel_velocities(vl, vr);
}

void 
carmen_pioneer_usage(char *progname, char *fmt, ...) 
{
  va_list args;

  if (fmt != NULL)
    {
      fprintf(stderr, "\n[31;1m");
      va_start(args, fmt);
      vfprintf(stderr, fmt, args);
      va_end(args);
      fprintf(stderr, "[0m\n\n");
    }
  else
    {
      fprintf(stderr, "\n");
    }

  if (strrchr(progname, '/') != NULL)
    {
      progname = strrchr(progname, '/');
      progname++;
    }
  
  fprintf(stderr, "Usage: pioneer <args>\n");
  fprintf(stderr, "\t-dev devname   - device connected to pioneer robot.\n");
  fprintf(stderr, "\t-version num  - version of the pioneer robot.\n");
  fprintf(stderr, "\t-sonar {on|off}  - use (or don't use) sonar.\n");
  exit(-1);
}



static int 
carmen_pioneer_initialize_ipc(void)
{
  IPC_RETURN_TYPE err;

  /* define messages created by this module */
  err = IPC_defineMsg(CARMEN_BASE_ODOMETRY_NAME, IPC_VARIABLE_LENGTH,
                      CARMEN_BASE_ODOMETRY_FMT);
  carmen_test_ipc_exit(err, "Could not define", CARMEN_BASE_ODOMETRY_NAME);

  err = IPC_defineMsg(CARMEN_BASE_SONAR_NAME, IPC_VARIABLE_LENGTH,
                      CARMEN_BASE_SONAR_FMT);
  carmen_test_ipc_exit(err, "Could not define", CARMEN_BASE_SONAR_NAME);

  err = IPC_defineMsg(CARMEN_BASE_VELOCITY_NAME, IPC_VARIABLE_LENGTH,
                      CARMEN_BASE_VELOCITY_FMT);
  carmen_test_ipc_exit(err, "Could not define", CARMEN_BASE_VELOCITY_NAME);

  /* setup incoming message handlers */
  err = IPC_subscribe(CARMEN_BASE_VELOCITY_NAME, velocity_handler, NULL);
  carmen_test_ipc_exit(err, "Could not subscribe", CARMEN_BASE_VELOCITY_NAME);
  IPC_setMsgQueueLength(CARMEN_BASE_VELOCITY_NAME, 1);

  return 0;
}

static int  
read_pioneer_parameters(int argc, char **argv) 
{
  int num_items;
  double wheel_diameter;
  carmen_param_t param_list[] = {
    {"pioneer", "dev", CARMEN_PARAM_STRING, &dev_name, 0, NULL},
    {"pioneer", "version", CARMEN_PARAM_INT, &pioneer_version, 0, NULL},
    {"pioneer", "wheel_diameter", CARMEN_PARAM_DOUBLE, &wheel_diameter, 0, NULL},
    {"pioneer", "relative_wheel_size", CARMEN_PARAM_DOUBLE, &pioneer_relative_wheel_size, 0, NULL},
    {"pioneer", "velocity_conversion_factor", CARMEN_PARAM_DOUBLE, 
     &pioneer_velocity_conversion_factor, 0, NULL},
    {"robot", "acceleration", CARMEN_PARAM_DOUBLE, &robot_acceleration, 1, NULL},
    {"robot", "deceleration", CARMEN_PARAM_DOUBLE, &robot_deceleration, 1, NULL},
    {"robot", "use_sonar", CARMEN_PARAM_ONOFF, &use_sonar, 1, NULL}  };
  carmen_param_t param_list_sonar[] = {
    {"robot", "max_t_vel", CARMEN_PARAM_DOUBLE, &robot_max_t_vel, 1, NULL},
    {"robot", "sensor_angle", CARMEN_PARAM_DOUBLE, &robot_sensor_angle, 1, NULL},
    {"robot", "sonar_offsets", CARMEN_PARAM_STRING, &robot_sonar_offsets, 1, NULL},
    {"robot", "num_sonars", CARMEN_PARAM_INT, &num_sonars, 1, NULL}  };

  num_items = sizeof(param_list)/sizeof(param_list[0]);
  carmen_param_install_params(argc, argv, param_list, num_items);
  if(use_sonar)
    {
      num_items = sizeof(param_list_sonar)/sizeof(param_list_sonar[0]);
      carmen_param_install_params(argc, argv, param_list_sonar, num_items);
    }
  if (pioneer_version == 1) 
    pioneer_distance_conversion_factor = M_PI * wheel_diameter / 
      PIONEER1_TICKS_PER_WHEEL_REVOLUTION;
  else
    pioneer_distance_conversion_factor = M_PI * wheel_diameter / 
      PIONEER2_TICKS_PER_WHEEL_REVOLUTION;

  if (robot_acceleration > robot_deceleration) 
    carmen_die("ERROR: robot_deceleration must be greater or equal than robot_acceleration\n");

  return 0;
}

int
carmen_pioneer_shutdown(int signo __attribute__ ((unused))) 
{
#ifdef _REENTRANT
  shutdown_in_progress = 1;
#endif
  fprintf(stderr, "\nShutting down robot...");
  sleep(1);
  pioneer_close();
  fprintf(stderr, "done.\n");
  return 0;
}

void
carmen_pioneer_emergency_crash(int signo __attribute__ ((unused)))
{
  command_vl = 0;
  command_vr = 0;
  set_velocity = 1;
  last_command = carmen_get_time_ms();
}

#ifdef _REENTRANT
static void *
start_thread(void *data __attribute__ ((unused))) 
{
  while (!shutdown_in_progress) {    
    pthread_testcancel();
    command_robot();
    usleep(30000);
  }
  thread_is_running = 0;
  return(NULL);
}  
#endif

int
carmen_pioneer_start(int argc, char **argv) 
{
  if (read_pioneer_parameters(argc, argv) < 0)
    return -1;

  if (carmen_pioneer_initialize_ipc() < 0) {
    fprintf(stderr, "\nError: Could not initialize IPC.\n");
    return -1;
  }

  if (carmen_serial_connect(&dev_fd, dev_name) < 0) 
    return -1;

  if (initialize_robot() < 0) {
    carmen_pioneer_shutdown(SIGINT);
    fprintf(stderr, "Error: Could not connect to robot.\n");
    return -1;
  }
  
  reset_status();

#ifdef _REENTRANT
  pthread_create(&main_thread, NULL, start_thread, NULL);
  thread_is_running = 1;
#endif

  return 0;
}

int
carmen_pioneer_run(void) 
{
  pioneer_pulse();
  update_status();
  publish_odometry_message(&odometry);
  if (use_sonar)
    {
      publish_sonar_message(&sonar);
    }
#ifndef _REENTRANT  
  command_robot();
#endif
  
  return 0;
}
