#include <carmen/carmen.h>
#include <netinet/in.h>
#include "rflex.h"
#include "rflex_commands.h"
#include "rflex-io.h"

typedef struct {
  int distance;
  int bearing;
  int t_vel;
  int r_vel;
  int num_sonars;
  int ranges[MAX_NUM_SONARS];
  int num_bumpers;
  char *bumpers;
} rflex_status_t;

static rflex_status_t status;

static int
sgn( long val )
{
  if (val<0) {
    return(0);
  } else {
    return(1);
  }
}

/* COMPUTE CRC CODE */

static int
computeCRC( unsigned char *buf, int nChars )
{ 
  int i, crc;
  if (nChars==0) {
    crc = 0;
  } else {
    crc = buf[0];
    for (i=1; i<nChars; i++) {
      crc ^= buf[i];
    }
  }
  return(crc);
}

/* CONVERSION BYTES -> NUM */

#if 0 

static unsigned int
convertBytes2UInt8( unsigned char *bytes )
{
  unsigned int i;
  memcpy( &i, bytes, 1 );
  return(i);
}

#endif

static unsigned int
convertBytes2UInt16( unsigned char *bytes )
{
  unsigned int i;
  memcpy( &i, bytes, 2 );
  return(htons(i));
}


static unsigned long
convertBytes2UInt32( unsigned char *bytes )
{
  unsigned long i;
  memcpy( &i, bytes, 4 );
  return(htonl(i));
}

/* CONVERSION NUM -> BYTES */

static void
convertUInt8( unsigned int i, unsigned char *bytes )
{
  memcpy( bytes, &i, 1 );
}

#if 0 

static void
convertUInt16( unsigned int i, unsigned char *bytes )
{
  uint16_t conv;
  conv = htonl( i );
  memcpy( bytes, &conv, 2 );
}

#endif

static void
convertUInt32( unsigned long l, unsigned char *bytes )
{
  uint32_t conv;
  conv = htonl( l );
  memcpy( bytes, &conv, 4 );
}

static void
cmdSend( int fd, int port, int id, int opcode, int len, unsigned char *data )
{
  int i;
  static unsigned char cmd[MAX_COMMAND_LENGTH];
  /* START CODE */
  cmd[0] = 0x1b;
  cmd[1] = 0x02;
  /* PORT */
  cmd[2] = (unsigned char) port;
  /* ID */
  cmd[3] = (unsigned char) id;
  /* OPCODE */
  cmd[4] = (unsigned char) opcode;
  /* LENGTH */
  cmd[5] = (unsigned char) len;
  /* DATA */
  for (i=0; i<len; i++) {
    cmd[6+i] = data[i];
  }
  /* CRC */
  cmd[6+len] = computeCRC( &(cmd[2]), len+4 );    /* END CODE */
  cmd[6+len+1] = 0x1b;
  cmd[6+len+2] = 0x03;

  writeData( fd, cmd, 9+len );
}

void
carmen_rflex_sonars_on( int fd )
{
  unsigned char data[MAX_COMMAND_LENGTH];
  convertUInt32( (long) 30000, &(data[0]) );
  convertUInt32( (long) 0, &(data[4]) );
  convertUInt32( (long) 0, &(data[8]) );
  convertUInt8(  (int) 2, &(data[12]) );
  cmdSend( fd, SONAR_PORT, 4, SONAR_RUN, 13, data );
}

void
carmen_rflex_sonars_off( int fd )
{
  unsigned char data[MAX_COMMAND_LENGTH];
  convertUInt32( (long) 0, &(data[0]) );
  convertUInt32( (long) 0, &(data[4]) );
  convertUInt32( (long) 0, &(data[8]) );
  convertUInt8(  (int) 0, &(data[12]) );
  cmdSend( fd, SONAR_PORT, 4, SONAR_RUN, 13, data );
}

void
carmen_rflex_digital_io_on( int fd, int period )
{
  unsigned char data[MAX_COMMAND_LENGTH];
  convertUInt32( (long) period, &(data[0]) );
  cmdSend( fd, DIO_PORT, 0, DIO_REPORTS_REQ, 4, data );
}

void
carmen_rflex_digital_io_off( int fd )
{
  unsigned char data[MAX_COMMAND_LENGTH];
  convertUInt32( (long) 0, &(data[0]) );
  cmdSend( fd, SONAR_PORT, 4, DIO_REPORTS_REQ, 4, data );
}

void
carmen_rflex_brake_on( int fd )
{
  cmdSend( fd, MOT_PORT, 0, MOT_BRAKE_SET, 0, NULL );
}

void
carmen_rflex_brake_off( int fd )
{
  cmdSend( fd, MOT_PORT, 0, MOT_BRAKE_RELEASE, 0, NULL );
}

void
carmen_rflex_motion_set_defaults( int fd )
{
  cmdSend( fd, MOT_PORT, 0, MOT_SET_DEFAULTS, 0, NULL );
}

void
carmen_rflex_odometry_on( int fd, long period )
{ 
  unsigned char data[MAX_COMMAND_LENGTH];
  convertUInt32( period, &(data[0]) );         /* period in ms */
  convertUInt32( (long) 3, &(data[4]) );       /* mask */
  cmdSend( fd, MOT_PORT, 0, MOT_SYSTEM_REPORT_REQ, 8, data );
}

void
carmen_rflex_odometry_off( int fd )
{ 
  unsigned char data[MAX_COMMAND_LENGTH];
  convertUInt32( (long) 0, &(data[0]) );       /* period in ms */
  convertUInt32( (long) 0, &(data[4]) );       /* mask */
  cmdSend( fd, MOT_PORT, 0, MOT_SYSTEM_REPORT_REQ, 8, data );
}

void
carmen_rflex_update_system( int fd , int *battery, 
			    int *timestamp, int *brake)
{
  cmdSend( fd, SYS_PORT, 0, SYS_STATUS, 0, NULL );
  
  battery = battery;
  timestamp = timestamp;
  brake = brake;
}

void
carmen_rflex_set_velocity( int fd, long tvel, long rvel,
			   long acceleration )
{
  unsigned char data[MAX_COMMAND_LENGTH];

  convertUInt8( (long) 0,                 &(data[0]) );       /* forward motion */
  convertUInt32( (long) abs(tvel),        &(data[1]) );       /* abs trans velocity */
  convertUInt32( (long) acceleration,    &(data[5]) );       /* trans acc */
  convertUInt32( (long) STD_TRANS_TORQUE, &(data[9]) );       /* trans torque */
  convertUInt8( (long) sgn(tvel),         &(data[13]) );      /* trans direction */

  cmdSend( fd, MOT_PORT, 0, MOT_AXIS_SET_DIR, 14, data );

  convertUInt8( (long) 1,                 &(data[0]) );       /* rotational motion */
  convertUInt32( (long) abs(rvel),        &(data[1]) );       /* abs rot velocity  */
  /* 0.275 rad/sec * 10000 */ 
  convertUInt32( (long) STD_ROT_ACC,      &(data[5]) );       /* rot acc */
  convertUInt32( (long) STD_ROT_TORQUE,   &(data[9]) );       /* rot torque */
  convertUInt8( (long) sgn(rvel),         &(data[13]) );      /* rot direction */

  cmdSend( fd, MOT_PORT, 0, MOT_AXIS_SET_DIR, 14, data );
}

void
carmen_rflex_stop_robot( int fd, int deceleration)
{
  carmen_rflex_set_velocity( fd, 0, 0, deceleration);
}

int
carmen_rflex_open_connection(char *device_name, int *fd)
{
  Device   rdev;

  strncpy( rdev.ttyport, device_name, MAX_NAME_LENGTH);
  rdev.baud           = 115200;
  rdev.databits       = 8;
  rdev.parity         = N;
  rdev.stopbits       = 1;
  rdev.hwf            = 0;
  rdev.swf            = 0;
  
  if (DEVICE_connect_port( &rdev )<0) {
    carmen_warn("Can't open device\n" );
    return -1;
  } 

  *fd = rdev.fd;

  carmen_rflex_odometry_on(*fd, 100000);
  carmen_rflex_digital_io_on(*fd, 100000);
  carmen_rflex_motion_set_defaults(*fd);

  return 0;
}

static void
parseMotReport( unsigned char *buffer )
{
  int rv, timeStamp, acc, trq;
  unsigned char axis, opcode;
    
  opcode = buffer[4];
  switch(opcode) {
  case MOT_SYSTEM_REPORT:
    rv        = convertBytes2UInt32(&(buffer[6]));
    timeStamp = convertBytes2UInt32(&(buffer[10]));
    axis      = buffer[14];
    if (axis == 0) {
      status.distance = convertBytes2UInt32(&(buffer[15]));
      status.t_vel = convertBytes2UInt32(&(buffer[19]));
    } else if (axis == 1) {
      status.bearing = convertBytes2UInt32(&(buffer[15]));
      status.r_vel = convertBytes2UInt32(&(buffer[19]));
    }
    acc       = convertBytes2UInt32(&(buffer[23]));
    trq       = convertBytes2UInt32(&(buffer[27]));
    break;
  default:
    break;
  }
}

static void
parseSonarReport( unsigned char *buffer )
{
  unsigned int sid;
 
  int count, retval, timeStamp;
  unsigned char opcode, dlen;
    
  opcode = buffer[4];
  dlen   = buffer[5];

  status.num_sonars=MAX_NUM_SONARS;
  switch(opcode) {
  case SONAR_REPORT:
    retval    = convertBytes2UInt32(&(buffer[6]));
    timeStamp = convertBytes2UInt32(&(buffer[10]));
    count = 0;
    while ((8+count*3<dlen) && (count<256)) {
      sid   = buffer[14+count*3];
      status.ranges[sid] = convertBytes2UInt16( &(buffer[14+count*3]) );
      count++;
    }
    //	    fprintf( stderr, "(s:%d)", count );
    break;
  default:
    break;
  }
}

static int
parseBuffer( unsigned char *buffer, unsigned int len )
{
  unsigned int port, dlen, crc;

  port   = buffer[2];
  dlen   = buffer[5];

  if (dlen+8>len) {
    return(0);
  } else {
    crc    = computeCRC( &(buffer[2]), dlen+4 );
    if (crc != buffer[len-3])
      return(0);
    switch(port) {
    case SYS_PORT:
      fprintf( stderr, "(sys)" );
      break;
    case MOT_PORT:
      parseMotReport( buffer );
      break;
    case JSTK_PORT:
      break;
    case SONAR_PORT:
      parseSonarReport( buffer );
      break;
    case DIO_PORT:
      fprintf( stderr, "(dio)" );
      break;
    case IR_PORT:
      fprintf( stderr, "(ir)" );
      break;
    default:
      break;
    }
  }
  return(1);
}

static void
clear_incoming_data(int fd)
{
  unsigned char buffer[4096];
  int len;
  int bytes;

  // 32 bytes here because the motion packet is 34. No sense in waiting for a
  // complete packet -- we can starve because the last 2 bytes might always
  // arrive with the next packet also, and we never leave the loop. 

  while ((bytes = bytesWaiting(fd)) > 32) {
    waitForAnswer(fd, buffer, &len);
    parseBuffer(buffer, len);
  }
}

void
carmen_rflex_update_status(int fd, int *distance,  int *bearing, 
			   int *t_vel, int *r_vel)
{
  clear_incoming_data(fd);

  *distance = status.distance;
  *bearing = status.bearing;
  *t_vel = status.t_vel;
  *r_vel = status.r_vel;
}

void
carmen_rflex_update_sonar(int fd, int num_sonars, 
			  int *ranges)
{
  int i;

  clear_incoming_data(fd);

  if (num_sonars > status.num_sonars) 
    {
      carmen_warn("Requested more sonars (%d) than available (%d).\n",
		  num_sonars, MAX_NUM_SONARS);
      num_sonars = status.num_sonars;
    }

  //copy all actual data

  for(i=0; i < num_sonars && i < status.num_sonars; i++)
    {
      if (status.ranges[i] < 0)
	continue;
      else if(status.ranges[i]==32766)
	ranges[i]=0;
      else
	ranges[i]=status.ranges[i];
    }

  if ( i < num_sonars)
    carmen_warn("Requested more sonars (%d) than available (%d).\n",
		num_sonars, i);
  
  //  memcpy(ranges, status.ranges, num_sonars*sizeof(int));
}

void
carmen_rflex_update_bumpers(int fd, int num_bumpers, 
			    char *values)
{
  clear_incoming_data(fd);

  if (num_bumpers > status.num_bumpers) {
    carmen_warn("Requested more bumpers than available.\n");
    num_bumpers = status.num_bumpers;
  }

  memcpy(values, status.bumpers, num_bumpers*sizeof(char));
}

void 
carmen_rflex_initialize(int fd, int trans_acceleration,
			int rot_acceleration,
			int trans_pos __attribute__ ((unused)),
			int rot_pos __attribute__ ((unused)))
{
  unsigned char data[MAX_COMMAND_LENGTH];
  int i;

  data[0] = 0;
  convertUInt32( (long) 0, &(data[1]) );          /* velocity */ 
  convertUInt32( (long) trans_acceleration, &(data[5]) );       
                                                  /* acceleration */ 
  convertUInt32( (long) 0, &(data[9]) );      /* torque */ 
  data[13] = 0;

  cmdSend( fd, MOT_PORT, 0, MOT_AXIS_SET_DIR, 14, data );

  data[0] = 1;
  convertUInt32( (long) 0, &(data[1]) );          /* velocity */ 
  convertUInt32( (long) rot_acceleration, &(data[5]) );       
                                                  /* acceleration */ 
  convertUInt32( (long) 0, &(data[9]) );      /* torque */ 
  data[13] = 0;

  cmdSend( fd, MOT_PORT, 0, MOT_AXIS_SET_DIR, 14, data );

  //mark all non-existant (or no-data) sonar as such
  //note - this varies from MAX_INT set when sonar fail to get a valid echo.

  for (i=0; i < MAX_NUM_SONARS; i++)
     status.ranges[i] = -1;
}
