/*********************************************************
 *
 * This source code is part of the Carnegie Mellon Robot
 * Navigation Toolkit (CARMEN)
 *
 * CARMEN Copyright (c) 2002 Michael Montemerlo, Nicholas
 * Roy, and Sebastian Thrun
 *
 * CARMEN is free software; you can redistribute it and/or 
 * modify it under the terms of the GNU General Public 
 * License as published by the Free Software Foundation; 
 * either version 2 of the License, or (at your option)
 * any later version.
 *
 * CARMEN is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied 
 * warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
 * PURPOSE.  See the GNU General Public License for more 
 * details.
 *
 * You should have received a copy of the GNU General 
 * Public License along with CARMEN; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place, 
 * Suite 330, Boston, MA  02111-1307 USA
 *
 ********************************************************/

#include <carmen/carmen.h>

#include <limits.h>

#include "rflex_commands.h"
#include "rflex_brake.h"
#include "rflex_params.h"

#ifdef _REENTRANT
#include <pthread.h>
static pthread_t main_thread;
static int thread_is_running = 0;
static int shutdown_in_progress = 0;
#endif

#define RFLEX_TIMEOUT 0.1

static char *rflex_dev;
static int rflex_fd;

static int set_velocity = 0;
static int command_t_vel, command_r_vel;
static double last_command = 0;
static int robot_moving = 0;

static double acceleration = 1.0;
static double deceleration = 1.0;

static carmen_base_odometry_message odometry;
static carmen_base_sonar_message sonar;
static carmen_base_bumper_message bumpers;
static carmen_base_ir_message ir;

static int use_ir = 0;

static int use_sonar = 0;
static int *ranges = NULL;

static double angle_conversion;
static double distance_conversion;


/*** TABLE OF CONTENTS ***/

/*** Higher-level robot functions ***/

static void  command_robot(void);
static int   initialize_robot(char *dev);
static int   read_rflex_parameters(int argc, char **argv);
static void  reset_status(void);
static void  update_status(void);

static void 
command_robot(void)
{
  double time_elapsed = carmen_get_time_ms() - last_command;

  if(set_velocity) 
    {
      if(command_t_vel == 0 && command_r_vel == 0) 
	{
	  fprintf(stderr, "S");
	  carmen_rflex_stop_robot(rflex_fd, deceleration*distance_conversion);
	  robot_moving = 0;
	}
      else 
	{
	  fprintf(stderr, "V");
	  carmen_rflex_set_velocity(rflex_fd, command_t_vel, command_r_vel,
				    acceleration*distance_conversion);
	  robot_moving = 1;
	  last_command = carmen_get_time_ms();
	}
      set_velocity = 0;
    }
  else if(robot_moving && time_elapsed > RFLEX_TIMEOUT)
    {
      fprintf(stderr, "T");
      carmen_rflex_stop_robot(rflex_fd, deceleration*distance_conversion);
      command_t_vel = 0;
      command_r_vel = 0;

      robot_moving = 0;
      set_velocity = 0;
    }
}


static int 
initialize_robot(char *dev) 
{
#ifdef _REENTRANT
  if (thread_is_running)
    {
      carmen_warn("Tried to connect to the robot while the command loop "
		  "thread is running.\n");
      carmen_warn("This is a bug in the code, and must be fixed.\n");
      return -1;
    }
#endif

  if (carmen_rflex_open_connection(dev, &rflex_fd) < 0) 
    return -1;

  if (use_sonar) {
    carmen_rflex_sonars_on(rflex_fd);
  }

  carmen_rflex_brake_off(rflex_fd);

  carmen_rflex_initialize(rflex_fd, acceleration*distance_conversion, 
			  M_PI/4 * angle_conversion, 0, 0);

  return 0;
}

static void 
reset_status(void) {
  odometry.x = 0.0;
  odometry.y = 0.0;
  odometry.theta = 0.0;
  odometry.tv = 0.0;
  odometry.rv = 0.0;
  odometry.timestamp = 0.0;
  
  if (use_sonar) 
    {
      sonar.timestamp = 0.0;
      strncpy(sonar.host, odometry.host, 10);
    }
}


static void 
update_status(void) 
{
  double timestamp;

  static int initialized = 0;
  static int last_range_position;
  static int last_bearing_position;
  static double x = 0, y = 0, theta =0;

  int new_range_position, new_bearing_position;
  int t_vel, r_vel;
  double displacement;

  int i;

  timestamp = carmen_get_time_ms();

  carmen_rflex_update_status(rflex_fd, &new_range_position,
			     &new_bearing_position, &t_vel,
			     &r_vel);

  if (!initialized) {
    initialized = 1;
  } else {
    
    /* This condition here is to trap when the angle odometry rolls over. */
    if (0 && fabs(new_bearing_position - last_bearing_position) > SHRT_MAX/2)
      {
	last_bearing_position -= SHRT_MAX;
	new_bearing_position += (INT_MAX-SHRT_MAX);	
      }

    /* This condition here is to trap when the translation odometry 
       rolls over. */
    if (0 && fabs(new_range_position - last_range_position) > SHRT_MAX/2)
      {
	last_range_position -= SHRT_MAX;
	new_range_position += (INT_MAX-SHRT_MAX);	
      }

    theta += (new_bearing_position - last_bearing_position) / angle_conversion;
    theta = carmen_normalize_theta(theta);
    displacement = (new_range_position - last_range_position) /
      distance_conversion;

    x += displacement * cos(theta);
    y += displacement * sin(theta);

    odometry.x = x;
    odometry.y = y;
    odometry.theta = theta;
    
    odometry.tv = t_vel / distance_conversion;
    odometry.rv = r_vel / angle_conversion;
    odometry.timestamp = carmen_get_time_ms();
  }

  last_range_position = new_range_position;
  last_bearing_position = new_bearing_position;

  if (use_sonar) 
    {
      carmen_rflex_update_sonar(rflex_fd, sonar.num_sonars,
				ranges);
      for (i = 0; i < sonar.num_sonars; i++) 
	sonar.range[i] = ranges[i] / 1000; // mm to meters
      
      sonar.timestamp = timestamp;  
    }

  carmen_rflex_update_bumpers(rflex_fd, bumpers.num_bumpers,
			      bumpers.state);
}

static void 
publish_odometry_message(carmen_base_odometry_message *odometry_msg)
{
  IPC_RETURN_TYPE err;
  char *host;
  static int first = 1;

  if(first) {
    host = carmen_get_tenchar_host_name();
    strcpy(odometry_msg->host, host);
    first = 0;
  }

  err = IPC_publishData(CARMEN_BASE_ODOMETRY_NAME, odometry_msg);
  carmen_test_ipc_exit(err, "Could not publish", CARMEN_BASE_ODOMETRY_NAME);
}

static void 
publish_sonar_message(carmen_base_sonar_message *sonar_msg)
{
  IPC_RETURN_TYPE err;
  char *host;
  static int first = 1;

  if(first) {
    host = carmen_get_tenchar_host_name();
    strcpy(sonar_msg->host, host);
    first = 0;
  }

  err = IPC_publishData(CARMEN_BASE_SONAR_NAME, sonar_msg);
  carmen_test_ipc_exit(err, "Could not publish", CARMEN_BASE_SONAR_NAME);
}

static void
publish_bumper_message (carmen_base_bumper_message *bumper_msg)
{
  IPC_RETURN_TYPE err;
  char *host;
  static int first = 1;

  if(first) {
    host = carmen_get_tenchar_host_name();
    strcpy(bumper_msg->host, host);
    first = 0;
  }

  err = IPC_publishData(CARMEN_BASE_BUMPER_NAME, bumper_msg);
  carmen_test_ipc_exit(err, "Could not publish", CARMEN_BASE_BUMPER_NAME);
}

static void
publish_ir_message (carmen_base_ir_message *ir_msg)
{
  IPC_RETURN_TYPE err;
  char *host;
  static int first = 1;

  if(first) {
    host = carmen_get_tenchar_host_name();
    strcpy(ir_msg->host, host);
    first = 0;
  }

  err = IPC_publishData(CARMEN_BASE_IR_NAME, ir_msg);
  carmen_test_ipc_exit(err, "Could not publish", CARMEN_BASE_IR_NAME);
}

static void 
velocity_handler(MSG_INSTANCE msgRef, BYTE_ARRAY callData,
		 void *clientData __attribute__ ((unused)))
{
  IPC_RETURN_TYPE err;
  FORMATTER_PTR formatter;

  carmen_base_velocity_message vel;
  
  formatter = IPC_msgInstanceFormatter(msgRef);
  err = IPC_unmarshallData(formatter, callData, &vel,
			   sizeof(carmen_base_velocity_message));
  IPC_freeByteArray(callData);
  carmen_test_ipc_return(err, "Could not unmarshall", 
			 IPC_msgInstanceName(msgRef));

  set_velocity = 1;
  command_t_vel = vel.tv*distance_conversion;
  command_r_vel = vel.rv*angle_conversion;
}

static void 
brake_handler(MSG_INSTANCE msgRef, BYTE_ARRAY callData,
	      void *clientData __attribute__ ((unused)))
{
  IPC_RETURN_TYPE err;
  FORMATTER_PTR formatter;

  carmen_rflex_brake_message data;
  
  formatter = IPC_msgInstanceFormatter(msgRef);
  err = IPC_unmarshallData(formatter, callData, &data,
			   sizeof(carmen_rflex_brake_message));
  IPC_freeByteArray(callData);
  carmen_test_ipc_return(err, "Could not unmarshall", 
			 IPC_msgInstanceName(msgRef));

  if (data.set_brake==0)
    carmen_rflex_brake_off(rflex_fd);
  else
    carmen_rflex_brake_on(rflex_fd);
}

void 
carmen_rflex_usage(char *progname, char *fmt, ...) 
{
  va_list args;

  if (fmt != NULL)
    {
      fprintf(stderr, "\n[31;1m");
      va_start(args, fmt);
      vfprintf(stderr, fmt, args);
      va_end(args);
      fprintf(stderr, "[0m\n\n");
    }
  else
    {
      fprintf(stderr, "\n");
    }

  if (strrchr(progname, '/') != NULL)
    {
      progname = strrchr(progname, '/');
      progname++;
    }
  
  fprintf(stderr, "Usage: rflex <args>\n");
  fprintf(stderr, "\t-dev devname   - device connected to rflex robot.\n");
  fprintf(stderr, "\t-version num  - version of the rflex robot.\n");
  fprintf(stderr, "\t-sonar {on|off}  - use (or don't use) sonar.\n");
  exit(-1);
}



static int 
carmen_rflex_initialize_ipc(void)
{
  IPC_RETURN_TYPE err;

  /* define messages created by this module */
  err = IPC_defineMsg(CARMEN_BASE_ODOMETRY_NAME, IPC_VARIABLE_LENGTH,
                      CARMEN_BASE_ODOMETRY_FMT);
  carmen_test_ipc_exit(err, "Could not define", CARMEN_BASE_ODOMETRY_NAME);

  err = IPC_defineMsg(CARMEN_BASE_SONAR_NAME, IPC_VARIABLE_LENGTH,
                      CARMEN_BASE_SONAR_FMT);
  carmen_test_ipc_exit(err, "Could not define", CARMEN_BASE_SONAR_NAME);

  err = IPC_defineMsg(CARMEN_BASE_BUMPER_NAME, IPC_VARIABLE_LENGTH,
                      CARMEN_BASE_BUMPER_FMT);
  carmen_test_ipc_exit(err, "Could not define", CARMEN_BASE_BUMPER_NAME);

  err = IPC_defineMsg(CARMEN_BASE_VELOCITY_NAME, IPC_VARIABLE_LENGTH,
                      CARMEN_BASE_VELOCITY_FMT);
  carmen_test_ipc_exit(err, "Could not define", CARMEN_BASE_VELOCITY_NAME);

  err = IPC_defineMsg(CARMEN_RFLEX_BRAKE_NAME, IPC_VARIABLE_LENGTH,
                      CARMEN_RFLEX_BRAKE_FMT);
  carmen_test_ipc_exit(err, "Could not define", CARMEN_RFLEX_BRAKE_NAME);

  /* setup incoming message handlers */
  err = IPC_subscribe(CARMEN_BASE_VELOCITY_NAME, velocity_handler, NULL);
  carmen_test_ipc_exit(err, "Could not subscribe", CARMEN_BASE_VELOCITY_NAME);
  IPC_setMsgQueueLength(CARMEN_BASE_VELOCITY_NAME, 1);

  err = IPC_subscribe(CARMEN_RFLEX_BRAKE_NAME, brake_handler, NULL);
  carmen_test_ipc_exit(err, "Could not subscribe", CARMEN_RFLEX_BRAKE_NAME);
  IPC_setMsgQueueLength(CARMEN_RFLEX_BRAKE_NAME, 1);

  return 0;
}

static void
start_sonars(void)
{
  double sonar_angle;
  char *offset_string;
  int num_sonars;

  carmen_param_set_module("robot");
  carmen_param_get_int("num_sonars", &num_sonars);
  carmen_param_get_double("sensor_angle", &sonar_angle);
  carmen_param_get_string("sonar_offsets", &offset_string);

  sonar.num_sonars = num_sonars;
  sonar.range = (double *)calloc(num_sonars, sizeof(double));
  carmen_test_alloc(sonar.range);

  ranges = (int *)calloc(num_sonars, sizeof(int));
  carmen_test_alloc(ranges);

  sonar.sonar_offsets = (carmen_point_p)calloc(num_sonars, 
					       sizeof(carmen_point_t));
  carmen_test_alloc(sonar.sonar_offsets);
  carmen_parse_sonar_offsets(offset_string, sonar.sonar_offsets, num_sonars);
  sonar.sensor_angle= sonar_angle;

  carmen_rflex_sonars_on(rflex_fd);
}

static void
handle_sonar_change(char *module __attribute__ ((unused)), 
		    char *variable __attribute__ ((unused)), 
		    char *value __attribute__ ((unused)))
{
  if (use_sonar)
    start_sonars();
  else {
    if (ranges)
      free(ranges);
    if (sonar.range)
      free(sonar.range);
    if (sonar.sonar_offsets)
      free(sonar.sonar_offsets);
    carmen_rflex_sonars_off(rflex_fd);
  }
}

static int  
read_rflex_parameters(int argc, char **argv) 
{
  int num_items;
  char *rflex_type;
  int rflex_model;
  int i;

  carmen_param_t param_list[] = {
    {"rflex", "type", CARMEN_PARAM_STRING, &rflex_type, 0, NULL},
    {"rflex", "dev", CARMEN_PARAM_STRING, &rflex_dev, 0, NULL},
    {"robot", "acceleration", CARMEN_PARAM_DOUBLE, &acceleration, 1, NULL},
    {"robot", "deceleration", CARMEN_PARAM_DOUBLE, &deceleration, 1, NULL},
    {"robot", "use_sonar", CARMEN_PARAM_ONOFF, &use_sonar, 1, 
    handle_sonar_change}};

  num_items = sizeof(param_list)/sizeof(param_list[0]);
  carmen_param_install_params(argc, argv, param_list, num_items);

  rflex_model = -1;
  for (i = 0; carmen_rflex_models[i][0] != 0; i++) 
    {
      if (strlen(rflex_type) == strlen(carmen_rflex_models[i]) &&
	  carmen_strncasecmp(rflex_type, carmen_rflex_models[i], 
			     strlen(carmen_rflex_models[i])) == 0)
	{
	  rflex_model = i;
	  break;
	}
  }

  if (rflex_model == -1) 
    {
      carmen_warn("Unknown rflex model %s\nKnown models: ", rflex_type);
      for (i = 0; carmen_rflex_models[i][0] != 0; i++) 
	{
	  carmen_warn("%s ", carmen_rflex_models[i]);
	  if (i % 6 == 5)
	    carmen_warn("\n");
	  carmen_warn("To determine your rflex model, read \n"
		      "carmen/src/rflex/README\n");
	  return -1;
	}
    }

  distance_conversion = carmen_rflex_params[rflex_model][0];
  angle_conversion = carmen_rflex_params[rflex_model][1];

  if(use_sonar)
    start_sonars();
  else {
    sonar.num_sonars = 0;
    sonar.range = NULL;
    sonar.sonar_offsets = NULL;
  }

  if (acceleration > deceleration) 
    carmen_die("ERROR: robot_deceleration must be greater or equal than "
	       "robot_acceleration\n");

  return 0;
}

int
carmen_rflex_shutdown(int signo __attribute__ ((unused))) 
{
#ifdef _REENTRANT
  shutdown_in_progress = 1;
#endif
  fprintf(stderr, "\nShutting down robot...");
  sleep(1);
  if (use_sonar)
    carmen_rflex_sonars_off(rflex_fd);
  fprintf(stderr, "done.\n");
  return 0;
}

void
carmen_rflex_emergency_crash(int signo __attribute__ ((unused)))
{
  command_t_vel = 0;
  command_r_vel = 0;
  set_velocity = 1;
  last_command = carmen_get_time_ms();
}

#ifdef _REENTRANT
static void *
start_thread(void *data __attribute__ ((unused))) 
{
  while (!shutdown_in_progress) {    
    pthread_testcancel();
    command_robot();
    usleep(30000);
  }
  thread_is_running = 0;
  return(NULL);
}  
#endif

int
carmen_rflex_start(int argc, char **argv) 
{
  if (read_rflex_parameters(argc, argv) < 0)
    return -1;

  if (carmen_rflex_initialize_ipc() < 0) {
    fprintf(stderr, "\nError: Could not initialize IPC.\n");
    return -1;
  }

  if (initialize_robot(rflex_dev) < 0) {
    carmen_rflex_shutdown(SIGINT);
    fprintf(stderr, "Error: Could not connect to robot.\n");
    return -1;
  }
  
  reset_status();

#ifdef _REENTRANT
  pthread_create(&main_thread, NULL, start_thread, NULL);
  thread_is_running = 1;
#endif

  return 0;
}

int
carmen_rflex_run(void) 
{
  update_status();
  publish_odometry_message(&odometry);
  publish_bumper_message(&bumpers);
  if (use_sonar)
    {
      publish_sonar_message(&sonar);
    }
  if (use_ir)
    {
      publish_ir_message(&ir);
    }
#ifndef _REENTRANT  
  command_robot();
#endif
  
  return 0;
}
