/*********************************************************
 *
 * This source code is part of the Carnegie Mellon Robot
 * Navigation Toolkit (CARMEN)
 *
 * CARMEN Copyright (c) 2002 Michael Montemerlo, Nicholas
 * Roy, and Sebastian Thrun
 *
 * CARMEN is free software; you can redistribute it and/or 
 * modify it under the terms of the GNU General Public 
 * License as published by the Free Software Foundation; 
 * either version 2 of the License, or (at your option)
 * any later version.
 *
 * CARMEN is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied 
 * warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
 * PURPOSE.  See the GNU General Public License for more 
 * details.
 *
 * You should have received a copy of the GNU General 
 * Public License along with CARMEN; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place, 
 * Suite 330, Boston, MA  02111-1307 USA
 *
 ********************************************************/

#include <carmen/carmen.h>

#include <sys/ioctl.h>
#include "Nclient.h"

#ifdef _REENTRANT
#include <pthread.h>
static pthread_t main_thread;
static int thread_is_running = 0;
#endif

#define        SCOUT_TIMEOUT          2.0

#define        METRES_PER_INCH        0.0254
#define        METRES_PER_SCOUT        (METRES_PER_INCH/10.0)
#define        DEGREES_PER_SCOUT       0.1

#define        WHEELBASE              (13.4 * METRES_PER_INCH)
#define        ROT_VEL_FACT_RAD       (WHEELBASE/METRES_PER_SCOUT)

#define        BACKWARDS_FLAG         1
#define        MAXV                   (1.0/METRES_PER_SCOUT)
#define        MAX_SONARS             16

static char *dev_name;
static int num_sonars = 16;
static int set_velocity = 0;
static int command_vl = 0, command_vr = 0;
static double last_command = 0;
static double last_update = 0;
static int timeout = 0;
static int moving = 0;
static double current_acceleration;
static double stopping_acceleration;
static carmen_robot_config_t robot_config;
static carmen_base_odometry_message odometry;
static carmen_base_sonar_message sonar;

static int use_sonar = 1;
static int scout_use_sonar=1;
static double sensor_angle;
static char* SONAR_OFFSETS;


static void parse_sonar_offsets(carmen_point_p sonar_offsets);

static int
turn_sonar_on(void)
{
  int sn_order[MAX_SONARS] = {0, 2, 15, 1, 14, 3, 13, 4, 12, 5, 11, 6, 10, 7, 9, 8};  
  int result;
  
  if (!use_sonar)
    return 0;

#ifdef _REENTRANT
  if (thread_is_running)
    {
      carmen_warn("Tried to call turn_sonar_on while the command loop thread is running.\n");
      carmen_warn("This is a bug in the code, and must be fixed. sonars not turned on or off.\n");
      return -1;
    }
#endif

  result = conf_sn(15, sn_order);
  if (result == FALSE)
    return -1;

  conf_tm(5); 
  if (result == FALSE)
    return -1;

  last_update = carmen_get_time_ms();
  fprintf(stderr, "getting closer\n");
  return 1;
}

static int
turn_sonar_off(void)
{
  if (!use_sonar)
    return 0;

  return turn_sonar_on();
}

static int 
initialize_robot(char *dev)
{
  int result;
  double acc;

#ifdef _REENTRANT
  if (thread_is_running)
    {
      carmen_warn("Tried to connect to the robot while the command loop thread is running.\n");
      carmen_warn("This is a bug in the code, and must be fixed. sonars not turned on.\n");
      return -1;
    }
#endif
  result = connect_robot(1, MODEL_SCOUT2, dev, 38400);
  if(result == FALSE)
    return -1;

  result = zr();
  if(result !=1)
    return -1;

  if (use_sonar) 
    {      
      result = turn_sonar_on();
      if(result!=1)
	return -1;
    }

  acc = robot_config.acceleration/METRES_PER_SCOUT;
  result = ac(acc, acc, 0.0);
  current_acceleration = robot_config.acceleration;

  if(result!=1)
    return -1;

  last_update = carmen_get_time_ms();

  return 1;
}

static void
parse_sonar_offsets(carmen_point_p sonar_offsets)
{
  char *so,*data, *curr, *next;
  int sonar_num=0, data_size=20;
  if(use_sonar)
    {
      so=(char*)calloc(strlen(SONAR_OFFSETS), sizeof(char));
      carmen_test_alloc(so);
      memcpy(so,SONAR_OFFSETS,strlen(SONAR_OFFSETS)*sizeof(char));
      data=(char*)calloc(data_size,sizeof(char));
      carmen_test_alloc(data);
      curr=so;
      while(curr!=NULL && sonar_num<num_sonars)
	{
	  next=strchr(curr, (int)' ');
	  memset(data,0,data_size);
	  if(next==NULL)
	    {
	      carmen_warn("Incorrect number of arguments in offset string. X\n");
	      break;
	    }
	  else
	    {
	      next[0]='\0';
	      data=strncpy(data,curr, strlen(curr)-strlen(next)+1);
	      if(data==NULL)
		{
		  carmen_warn("Too few arguments in offset string. X\n");
		  break;
		}
	      sonar_offsets[sonar_num].x=atof(data);
	      curr=&next[1];
	    }
	  
	  next=strchr(curr, (int)' ');
	  memset(data,0,data_size);
	  if(next==NULL)
	    {
	      carmen_warn("Incorrect number of arguments in offset string. Y\n");
	      break;
	    }
	  else
	    {
	      next[0]='\0';
	      data=strncpy(data,curr, strlen(curr)-strlen(next)+1);
	      if(data==NULL)
		{
		  carmen_warn("Too few arguments in offset string. Y\n");
		  break;
		}
	      sonar_offsets[sonar_num].y=atof(data);
	      curr=&next[1];
	    }
	  
	  next=strchr(curr, (int)' ');
	  memset(data, 0, data_size);
	  if(next==NULL)
	    {
	      data=strcpy(data,curr);
	      data[strlen(curr)]='\0';
	      if(data==NULL)
		{
		  carmen_warn("Too few arguments in offset string. THETA\n");
		  break;
		}
	      sonar_offsets[sonar_num].theta=carmen_normalize_theta(atof(data));
	      curr=NULL;
	    }
	  else
	    {
	      next[0]='\0';
	      data=strncpy(data,curr, strlen(curr)-strlen(next)+1);
	      if(data==NULL)
		{
		  carmen_warn("Too few arguments in offset string. Y\n");
		  break;
		}
	      sonar_offsets[sonar_num].theta=carmen_normalize_theta(atof(data));
	      curr=&next[3];
	    }
	  sonar_num++;
	}
      if(sonar_num<(num_sonars-1))
	carmen_warn("Too few arguments in offset string.  END\n");
    }
}

static void 
reset_status(void)
{
  char *host;
  odometry.x = 0.0;
  odometry.y = 0.0;
  odometry.theta = 0.0;
  odometry.tv = 0.0;
  odometry.rv = 0.0;
  odometry.timestamp = 0.0;

  host = carmen_get_tenchar_host_name();
  strcpy(odometry.host, host);
   
  if (use_sonar) 
    {
      sonar.num_sonars = num_sonars;
      sonar.range = (double *)calloc(num_sonars, sizeof(double));
      carmen_test_alloc(sonar.range);
      sonar.sonar_offsets = (carmen_point_p)calloc(num_sonars, sizeof(carmen_point_t));
      carmen_test_alloc(sonar.sonar_offsets);
      parse_sonar_offsets(sonar.sonar_offsets);
      sonar.sensor_angle=sensor_angle;
      sonar.timestamp = 0.0;
      strncpy(sonar.host, odometry.host, 10);
    }
}

static int
update_status(void)  /* This function takes approximately 60 ms to run */
{
  int i;
  double vl, vr;
  static double last_published_update;

  if (last_published_update >= odometry.timestamp)
    return 0; 

  /* update odometry message */
  
  odometry.x = ((double)State[STATE_CONF_X]) * METRES_PER_SCOUT;
  odometry.y = ((double)State[STATE_CONF_Y]) * METRES_PER_SCOUT;
  odometry.theta  = ((double)State[STATE_CONF_STEER]) * DEGREES_PER_SCOUT;

  odometry.theta = carmen_degrees_to_radians(odometry.theta);
  if(!BACKWARDS_FLAG) 
    {
      vl = ((double)State[STATE_VEL_TRANS]) * METRES_PER_SCOUT;
      vr = ((double)State[STATE_VEL_STEER]) * METRES_PER_SCOUT;
    }
  else 
    {
      odometry.theta = carmen_normalize_theta(odometry.theta+M_PI);      
      vl = -((double)State[STATE_VEL_STEER]) * METRES_PER_SCOUT;
      vr = -((double)State[STATE_VEL_TRANS]) * METRES_PER_SCOUT;
    }
  odometry.tv = 0.5 * (vl + vr);
  odometry.rv = (vl - vr) / ROT_VEL_FACT_RAD;
  
  /* update sonar message */
  if(use_sonar)
    {
      if(!(BACKWARDS_FLAG))
	for(i = 0; i < num_sonars; i++)
	  sonar.range[i] = ((double)State[i + STATE_SONAR_0]) * METRES_PER_INCH;
      else
	for(i = 0; i < num_sonars; i++)
	  sonar.range[(i + (num_sonars / 2)) % (num_sonars)]
	    = ((double) State[i + STATE_SONAR_0]) * METRES_PER_INCH;
      sonar.num_sonars=num_sonars;
    }
  last_published_update = odometry.timestamp;

  return 1;
}

static void 
command_robot(void)
{
  double time_elapsed = carmen_get_time_ms() - last_command;
  double current_time; 
  double acc;

  if (set_velocity)
    {
      if(command_vl == 0 && command_vr == 0) 
	{
	  moving = 0;
	  carmen_warn("S");
	}
      else 
	{
	  if (!moving)
	    {
	      acc = robot_config.acceleration/METRES_PER_SCOUT;
	      ac(acc, acc, 0.0);
	      current_acceleration = robot_config.acceleration;
	    }
	  carmen_warn("V");
	  moving = 1;
	}

      set_velocity = 0;
      timeout = 0;

      vm(command_vl, command_vr, 0.0);
      last_command = carmen_get_time_ms();
    }
  else if(time_elapsed > SCOUT_TIMEOUT && !timeout) 
    {
      carmen_warn("T");
      command_vl = 0;
      command_vr = 0;
      timeout = 1;
      moving = 0;
      vm(0.0, 0.0, 0.0);
    }  
  else if (moving && current_acceleration == robot_config.acceleration)
    {
      acc = stopping_acceleration/METRES_PER_SCOUT;
      ac(acc, acc, 0.0);
      current_acceleration = stopping_acceleration;
    }
  else
    {
      gs();
    }
  
  current_time = carmen_get_time_ms();
  last_update = current_time;
  odometry.timestamp = current_time;
  if(use_sonar)
    sonar.timestamp = current_time;
}

static int
read_scout_parameters(int argc, char **argv)
{
  int num_items;

  carmen_param_t param_list[] = {
    {"scout", "dev", CARMEN_PARAM_STRING, &dev_name, 0, NULL},
    {"robot", "use_sonar", CARMEN_PARAM_ONOFF, &use_sonar, 1, NULL}, 
    {"robot", "max_t_vel", CARMEN_PARAM_DOUBLE, &(robot_config.max_t_vel), 1, NULL},
    {"robot", "max_r_vel", CARMEN_PARAM_DOUBLE, &(robot_config.max_r_vel), 1, NULL},
    {"robot", "acceleration", CARMEN_PARAM_DOUBLE, &(robot_config.acceleration), 1, NULL},
    {"robot", "deceleration", CARMEN_PARAM_DOUBLE, &(stopping_acceleration), 1, NULL}};
  carmen_param_t param_list_sonar[] = {
    {"robot", "num_sonars", CARMEN_PARAM_ONOFF, &num_sonars, 1, NULL},
    {"robot", "sensor_angle", CARMEN_PARAM_DOUBLE, & sensor_angle, 1, NULL},
    {"robot", "sonar_offsets", CARMEN_PARAM_STRING, &SONAR_OFFSETS, 1, NULL} };

  num_items = sizeof(param_list)/sizeof(param_list[0]);
  carmen_param_install_params(argc, argv, param_list, num_items);
  use_sonar=use_sonar&scout_use_sonar;
  if(use_sonar)
    {
      num_items = sizeof(param_list_sonar)/sizeof(param_list_sonar[0]);
      carmen_param_install_params(argc, argv, param_list_sonar, num_items);
    }
  if (robot_config.acceleration > stopping_acceleration) 
    carmen_die("ERROR: robot_deceleration must be greater or equal than robot_acceleration\n");
  
  return 0;
}

static void 
velocity_handler(MSG_INSTANCE msgRef, BYTE_ARRAY callData,
		 void *clientData __attribute__ ((unused)))
{
  IPC_RETURN_TYPE err;
  carmen_base_velocity_message vel;
  double vl, vr;

  FORMATTER_PTR formatter;
  
  formatter = IPC_msgInstanceFormatter(msgRef);
  err = IPC_unmarshallData(formatter, callData, &vel,
                     sizeof(carmen_base_velocity_message));
  IPC_freeByteArray(callData);

  carmen_test_ipc_return(err, "Could not unmarshall", IPC_msgInstanceName(msgRef));

  vl = vel.tv / METRES_PER_SCOUT;
  vr = vel.tv / METRES_PER_SCOUT;  

  vl -= 0.5 * vel.rv * ROT_VEL_FACT_RAD;
  vr += 0.5 * vel.rv * ROT_VEL_FACT_RAD;

  if(vl > MAXV)
    vl = MAXV;
  else if(vl < -MAXV)
    vl = -MAXV;
  if(vr > MAXV)
      vr = MAXV;
  else if(vr < -MAXV)
    vr = -MAXV;

  if(BACKWARDS_FLAG) 
    {
      vl *= -1;
      vr *= -1;
    }

  command_vl = (int)vl;
  command_vr = (int)vr;

  set_velocity = 1;
}

int 
carmen_scout_initialize_ipc(void)
{
  IPC_RETURN_TYPE err;

  /* define messages created by scout */
  err = IPC_defineMsg(CARMEN_BASE_ODOMETRY_NAME, IPC_VARIABLE_LENGTH,
                      CARMEN_BASE_ODOMETRY_FMT);
  carmen_test_ipc_exit(err, "Could not define", CARMEN_BASE_ODOMETRY_NAME);

  err = IPC_defineMsg(CARMEN_BASE_SONAR_NAME, IPC_VARIABLE_LENGTH,
                      CARMEN_BASE_SONAR_FMT);
  carmen_test_ipc_exit(err, "Could not define IPC message", CARMEN_BASE_SONAR_NAME);

  err = IPC_defineMsg(CARMEN_BASE_VELOCITY_NAME, IPC_VARIABLE_LENGTH,
                      CARMEN_BASE_VELOCITY_FMT);
  carmen_test_ipc_exit(err, "Could not define", CARMEN_BASE_VELOCITY_NAME);

  /* setup incoming message handlers */

  err = IPC_subscribe(CARMEN_BASE_VELOCITY_NAME, velocity_handler, NULL);
  carmen_test_ipc_exit(err, "Could not subscribe", CARMEN_BASE_VELOCITY_NAME);
  IPC_setMsgQueueLength(CARMEN_BASE_VELOCITY_NAME, 1);

  return IPC_No_Error;
}

#ifdef _REENTRANT
static void *
start_thread(void *data __attribute__ ((unused))) 
{
  while (1) {    
    command_robot();
    usleep(30000);
  }
  return(NULL);
}  
#endif


int 
carmen_scout_start(int argc, char **argv)
{
  if (read_scout_parameters(argc, argv) < 0)
    return -1;

  if (carmen_scout_initialize_ipc() < 0) 
    {
      carmen_warn("\nError: Could not initialize IPC.\n");
      return -1;
    }

  if(initialize_robot(dev_name) < 0)
    {
      carmen_warn("\nError: Could not connect to robot. Did you remember to turn the base on?\n");
      return -1;
    }

  reset_status();

  strcpy(odometry.host, carmen_get_tenchar_host_name());

  fprintf(stderr, "after setting up the sonar\n");
#ifdef _REENTRANT
  pthread_create(&main_thread, NULL, start_thread, NULL);
  thread_is_running = 1;
#endif
  return 0;
}

int 
carmen_scout_run(void) 
{
  IPC_RETURN_TYPE err;

  if (update_status() == 1)
    {
      err = IPC_publishData(CARMEN_BASE_ODOMETRY_NAME, &odometry);
      carmen_test_ipc_exit(err, "Could not publish", CARMEN_BASE_ODOMETRY_NAME);
      
      if (use_sonar)
	{
	  err = IPC_publishData(CARMEN_BASE_SONAR_NAME, &sonar);
	  carmen_test_ipc_exit(err, "Could not publish", CARMEN_BASE_SONAR_NAME);
	}
    }
#ifndef _REENTRANT
  command_robot();
#endif
  
  return 1;
}

void 
carmen_scout_shutdown(int x)
{
  if(x == SIGINT) 
    {
      carmen_verbose("\nShutting down robot...");
      sleep(1);
      st();
      turn_sonar_off();
      lp();
      last_update = carmen_get_time_ms();
      disconnect_robot(1);
      carmen_verbose("done.\n");
    }
}

void 
carmen_scout_emergency_crash(int x __attribute__ ((unused)))
{
  vm(0.0, 0.0, 0.0);
  last_update = carmen_get_time_ms();
}
